import sys
if sys.version_info >= (3, 8):
    from typing import Callable, Literal, List, Optional, Any
else:
    from typing import Callable, List, Optional, Any
    from typing_extensions import Literal

from pydantic import BaseModel
from aiortc import RTCIceServer, RTCRtpSender, RTCRtpReceiver, RTCDataChannel, MediaStreamTrack
from .transport import IceCandidate, IceParameters, DtlsParameters
from ..ortc import ExtendedRtpCapabilities
from ..sctp_parameters import SctpParameters, SctpStreamParameters
from ..producer import ProducerCodecOptions
from ..rtp_parameters import RtpCodecCapability, RtpParameters, MediaKind, RtpEncodingParameters


class HandlerRunOptions(BaseModel):
    direction: Literal['send', 'recv']
    iceParameters: IceParameters
    iceCandidates: List[IceCandidate]
    dtlsParameters: DtlsParameters
    sctpParameters: Optional[SctpParameters]
    iceServers: Optional[RTCIceServer]
    iceTransportPolicy: Optional[Literal['all', 'relay']]
    additionalSettings: Optional[Any]
    proprietaryConstraints: Optional[Any]
    extendedRtpCapabilities: ExtendedRtpCapabilities

    class Config:
        arbitrary_types_allowed=True

class HandlerSendOptions(BaseModel):
    track: MediaStreamTrack
    encodings: List[RtpEncodingParameters] = []
    codecOptions: Optional[ProducerCodecOptions]
    codec: Optional[RtpCodecCapability]

    class Config:
        arbitrary_types_allowed=True

class HandlerSendResult(BaseModel):
    localId: str
    rtpParameters: RtpParameters
    rtpSender: Optional[RTCRtpSender]

    class Config:
        arbitrary_types_allowed=True

class HandlerReceiveOptions(BaseModel):
    trackId: str
    kind: MediaKind
    rtpParameters: RtpParameters

class HandlerReceiveResult(BaseModel):
    localId: str
    track: MediaStreamTrack
    rtpReceiver: Optional[RTCRtpReceiver]

    class Config:
        arbitrary_types_allowed=True

class HandlerSendDataChannelResult(BaseModel):
    dataChannel: RTCDataChannel
    sctpStreamParameters: SctpStreamParameters

    class Config:
        arbitrary_types_allowed=True

class HandlerReceiveDataChannelOptions(BaseModel):
    sctpStreamParameters: SctpStreamParameters
    label: Optional[str]
    protocol: Optional[str]

class HandlerReceiveDataChannelResult(BaseModel):
    dataChannel: RTCDataChannel

    class Config:
        arbitrary_types_allowed=True


import asyncio
from abc import ABCMeta, abstractmethod

from .mediasoup_listener import MediasoupListener
from .mediasoup_signaler import MessageType, Notification
from .room_peer import Peer, PeerAppData


class QueuedNotificationListener(MediasoupListener, metaclass=ABCMeta):

    def __init__(self, mePeer: Peer):
        super(QueuedNotificationListener, self).__init__(mePeer)
        self._notificationQueue: asyncio.Queue = None

    def setQueue(self, notificationQueue):
        self._notificationQueue = notificationQueue

    async def enqueue(self, message):
        # print('equeue: ', end=',')
        # print(asyncio.get_running_loop())
        await self._notificationQueue.put(Notification(None, message['method'], message['data']))

    async def dequeue(self):
        # print('dequeue: ', end=',')
        # print(asyncio.get_running_loop())
        return await self._notificationQueue.get()

    # for statistic and monitor purpose
    def queueSize(self):
        return self._notificationQueue.qsize()

    def resetQueue(self, notificationQueue):
        self._notificationQueue = notificationQueue

    @abstractmethod
    async def runLoop(self, notificationQueue: asyncio.Queue):
        pass


class BandwidthNotificationListener(QueuedNotificationListener):
    def __init__(self, mePeer: Peer):
        super(BandwidthNotificationListener, self).__init__(mePeer)

    async def runLoop(self, notificationQueue: asyncio.Queue):
        super().setQueue(notificationQueue)
        # print('runLoop: ', end=',')
        # print(asyncio.get_running_loop())
        while True:
            message = await self.dequeue()
            if message.method == MessageType.SERVER_NOTIFICATION_downlinkBwe.value:
                await self.onDownlinkBwe(message)
            else:
                pass

    # high frequency
    async def onDownlinkBwe(self, message: Notification):
        # print(f'onDownlinkBwe, message.data:{message.data}')
        pass


class PeerNotificationListener(QueuedNotificationListener):
    def __init__(self, mePeer: Peer):
        super(PeerNotificationListener, self).__init__(mePeer)

    async def runLoop(self, notificationQueue: asyncio.Queue):
        super().setQueue(notificationQueue)
        # print('runLoop: ', end=',')
        # print(asyncio.get_running_loop())
        while True:
            message = await self.dequeue()
            if message.method == MessageType.SERVER_NOTIFICATION_activeSpeaker.value:
                speakPeer = self.mePeer.room.getPeerByPeerId(message.data['peerId'])
                if message.data['peerId'] is not None:
                    speakPeer = self.mePeer.room.getPeerByPeerId(message.data['peerId'])
                await self.onActiveSpeaker(message, speakPeer)
            elif message.method == MessageType.SERVER_NOTIFICATION_newPeer.value:
                newPeer = self.mePeer.room.addPeer(peerId=message.data['id'],
                                                   data=PeerAppData(displayName=message.data['displayName'],
                                                                    device=message.data['device']))
                await self.onNewPeer(message, newPeer)
            elif message.method == MessageType.SERVER_NOTIFICATION_peerDisplayNameChanged.value:
                otherPeer = self.mePeer.room.getPeerByPeerId(message.data['peerId'])
                otherPeer.data.displayName = message.data['displayName']
                await self.onPeerDisplayNameChanged(message, otherPeer)
            elif message.method == MessageType.SERVER_NOTIFICATION_peerClosed.value:
                removedPeer = self.mePeer.room.removePeer(message.data['peerId'])
                await self.onPeerClosed(message, removedPeer)
            else:
                pass

    async def onNewPeer(self, message: Notification, newPeer: Peer):
        pass

    # high frequency
    async def onActiveSpeaker(self, message: Notification, speakPeer: Peer):
        pass

    async def onPeerDisplayNameChanged(self, message: Notification, otherPeer: Peer):
        pass

    async def onPeerClosed(self, message: Notification, removedPeer: Peer):
        pass


class ProducerNotificationListener(QueuedNotificationListener):
    def __init__(self, mePeer: Peer):
        super(ProducerNotificationListener, self).__init__(mePeer)

    async def runLoop(self, notificationQueue: asyncio.Queue):
        super().setQueue(notificationQueue)
        # print('runLoop: ', end='')
        # print(asyncio.get_running_loop())
        while True:
            message = await self.dequeue()
            if message.method == MessageType.SERVER_NOTIFICATION_producerScore.value:
                await self.onProducerScore(message)
            else:
                pass

    # high frequency
    async def onProducerScore(self, message: Notification):
        pass


class ConsumerNotificationListener(QueuedNotificationListener):
    def __init__(self, mePeer: Peer):
        super(ConsumerNotificationListener, self).__init__(mePeer)

    async def runLoop(self, notificationQueue: asyncio.Queue):
        super().setQueue(notificationQueue)
        # print('runLoop: ', end='')
        # print(asyncio.get_running_loop())
        while True:
            message = await self.dequeue()
            otherPeer = self.mePeer.room.getPeerByConsumerId(message.data['consumerId'])
            if message.method == MessageType.SERVER_NOTIFICATION_consumerScore.value:
                await self.onConsumerScore(message, otherPeer)
            elif message.method == MessageType.SERVER_NOTIFICATION_consumerLayersChanged.value:
                await self.onConsumerLayersChanged(message, otherPeer)
            elif message.method == MessageType.SERVER_NOTIFICATION_consumerPaused.value:
                await self.onConsumerPaused(message, otherPeer)
            elif message.method == MessageType.SERVER_NOTIFICATION_consumerResumed.value:
                await self.onConsumerResumed(message, otherPeer)
            elif message.method == MessageType.SERVER_NOTIFICATION_consumerClosed.value:
                await self.onConsumerClosed(message, otherPeer)
            else:
                pass

    async def onConsumerLayersChanged(self, message: Notification, otherPeer: Peer):
        pass

    # high frequency
    async def onConsumerScore(self, message: Notification, otherPeer: Peer):
        pass

    async def onConsumerPaused(self, message: Notification, otherPeer: Peer):
        pass

    async def onConsumerResumed(self, message: Notification, otherPeer: Peer):
        pass

    async def onConsumerClosed(self, message: Notification, otherPeer: Peer):
        pass


class DataConsumerNotificationListener(QueuedNotificationListener):
    def __init__(self, mePeer: Peer):
        super(DataConsumerNotificationListener, self).__init__(mePeer)

    async def runLoop(self, notificationQueue: asyncio.Queue):
        super().setQueue(notificationQueue)
        # print('runLoop: ', end='')
        # print(asyncio.get_running_loop())
        while True:
            message = await self.dequeue()
            if message.method == MessageType.SERVER_NOTIFICATION_dataConsumerClosed.value:
                otherPeer = self.mePeer.room.getPeerByDataConsumerId(message.data['dataConsumerId'])
                await self.onDataConsumerClosed(message, otherPeer)
            else:
                pass

    async def onDataConsumerClosed(self, message: Notification, otherPeer: Peer):
        pass

    def onMessage(self, otherPeer: Peer, message, label, protocol, appData):
        # print(f'DataChannel from {otherPeer}, {label}-{protocol}: {message}, appData:{appData}')
        pass



import asyncio.coroutines

from .mediasoup_listener import MediasoupListener
from .mediasoup_signaler import Request
from .room_peer import Peer


class ConsumerRequestListener(MediasoupListener):
    def __init__(self, mePeer: Peer):
        super(ConsumerRequestListener, self).__init__(mePeer)

    async def onNewConsumer(self, doSignalerResponse: asyncio.coroutines, message: Request, otherPeer: Peer):
        await doSignalerResponse
        pass


class DataConsumerRequestListener(MediasoupListener):
    def __init__(self, mePeer: Peer):
        super(DataConsumerRequestListener, self).__init__(mePeer)

    async def onNewDataConsumer(self, doSignalerResponse: asyncio.coroutines, message: Request, otherPeer: Peer):
        await doSignalerResponse
        pass



import sys
if sys.version_info >= (3, 8):
    from typing import Optional, Literal, List, Any, Callable, Dict
else:
    from typing import Optional, List, Any, Callable, Dict
    from typing_extensions import Literal

from enum import IntEnum
from aiortc import RTCIceServer
from pydantic import BaseModel
from ..ortc import ExtendedRtpCapabilities
from ..sctp_parameters import SctpParameters, SctpStreamParameters


class IceParameters(BaseModel):
    # ICE username fragment.
    usernameFragment: str
    # ICE password.
    password: str
    # ICE Lite.
    iceLite: Optional[bool]

class IceCandidate(BaseModel):
    # Unique identifier that allows ICE to correlate candidates that appear on
    # multiple transports.
    foundation: str
    # The assigned priority of the candidate.
    priority: int
    # The IP address of the candidate.
    ip: str
    # The protocol of the candidate.
    protocol: Literal['udp', 'tcp']
    # The port for the candidate.
    port: int
    # The type of candidate..
    type: Literal['host', 'srflx', 'prflx', 'relay']
    # The type of TCP candidate.
    tcpType: Optional[Literal['active', 'passive', 'so']]

# The hash function algorithm (as defined in the "Hash function Textual Names"
# registry initially specified in RFC 4572 Section 8) and its corresponding
# certificate fingerprint value (in lowercase hex string as expressed utilizing
# the syntax of "fingerprint" in RFC 4572 Section 5).
class DtlsFingerprint(BaseModel):
    algorithm: str
    value: str

DtlsRole = Literal['auto', 'client', 'server']

class DtlsParameters(BaseModel):
    # DTLS role. Default 'auto'.
    role: DtlsRole = 'auto'
    fingerprints: List[DtlsFingerprint]

ConnectionState = Literal['new', 'connecting', 'connected', 'failed', 'disconnected', 'closed']

class IpVersion(IntEnum):
    ipv4 = 4
    ipv6 = 6

class PlainRtpParameters(BaseModel):
    ip: str
    ipVersion: IpVersion
    port: int

class TransportOptions(BaseModel):
    id: str
    iceParameters: IceParameters
    iceCandidates: List[IceCandidate]
    dtlsParameters: DtlsParameters
    sctpParameters: Optional[SctpParameters]
    iceServers: Optional[List[RTCIceServer]]
    iceTransportPolicy: Optional[Literal['all', 'relay']]
    additionalSettings: Optional[dict] = None
    proprietaryConstraints: Any = None
    appData: Optional[dict] = {}

    class Config:
        arbitrary_types_allowed=True

class InternalTransportOptions(TransportOptions):
    direction: Literal['send', 'recv']
    handlerFactory: Callable
    extendedRtpCapabilities: Optional[ExtendedRtpCapabilities] = None
    canProduceByKind: Dict[str, bool]



from .core import *

class h264_profile_level_id:
    @staticmethod
    def isSameProfile(params1={}, params2={}):
        return isSameProfile(params1,params2)

    @staticmethod
    def generateProfileLevelIdForAnswer(local_supported_params={},remote_offered_params={}):
        return generateProfileLevelIdForAnswer(local_supported_params,remote_offered_params)



import sys
if sys.version_info >= (3, 8):
    from typing import Optional, Any, Literal
else:
    from typing import Optional, Any
    from typing_extensions import Literal

import logging
from pydantic import BaseModel
from pyee import AsyncIOEventEmitter
from aiortc import RTCDataChannel
from .emitter import EnhancedEventEmitter
from .sctp_parameters import SctpStreamParameters


class DataConsumerOptions(BaseModel):
    id: str
    dataProducerId: str
    sctpStreamParameters: SctpStreamParameters
    label: Optional[str]
    protocol: Optional[str]
    appData: Optional[dict] = {}

class DataConsumer(EnhancedEventEmitter):
    def __init__(
        self,
        id: str,
        dataProducerId: str,
        dataChannel: RTCDataChannel,
        sctpStreamParameters: SctpStreamParameters,
        appData: Optional[dict] = {},
        loop=None
    ):
        super(DataConsumer, self).__init__(loop=loop)

        # Closed flag.
        self._closed: bool = False
        # Observer instance.
        self._observer: AsyncIOEventEmitter = AsyncIOEventEmitter()

        self._id = id
        self._dataProducerId = dataProducerId
        self._dataChannel = dataChannel
        self._sctpStreamParameters = sctpStreamParameters
        self._appData = appData

        self._handleDataChannel()
    
    # DataConsumer id.
    @property
    def id(self) -> str:
        return self._id
    
    # Associated DataProducer id.
    @property
    def dataProducerId(self) -> str:
        return self._dataProducerId
    
    # Whether the DataConsumer is closed.
    @property
    def closed(self) -> bool:
        return self._closed
    
    # SCTP stream parameters.
    @property
    def sctpStreamParameters(self) -> SctpStreamParameters:
        return self._sctpStreamParameters
    
    # DataChannel readyState.
    @property
    def readyState(self) -> Literal["closed", "closing", "connecting", "open"]:
        return self._dataChannel.readyState
    
    # DataChannel label.
    @property
    def label(self) -> str:
        return self._dataChannel.label
    
    # DataChannel protocol.
    @property
    def protocol(self) -> str:
        return self._dataChannel.protocol

    # DataChannel binaryType.
    @property
    def binaryType(self) -> str:
        return self._dataChannel.binaryType
    
    @binaryType.setter
    def binaryType(self, binaryType: str):
        self._dataChannel.binaryType = binaryType
    
    # App custom data.
    @property
    def appData(self) -> Any:
        return self._appData
    
    # Invalid setter.
    @appData.setter
    def appData(self, value):
        raise Exception('cannot override appData object')
    
    # Observer.
    @property
    def observer(self) -> AsyncIOEventEmitter:
        return self._observer
    
    # Closes the DataConsumer.
    async def close(self):
        if self._closed:
            return
        
        logging.debug('DataConsumer close()')

        self._closed = True

        self._dataChannel.close()

        await self.emit_for_results('@close')

        # Emit observer event.
        self._observer.emit('close')
    
    # Transport was closed.
    def transportClosed(self):
        if self._closed:
            return

        logging.debug('DataConsumer transportClosed()')

        self._closed = True

        self._dataChannel.close()

        self.emit('transportclose')

        self._observer.emit('close')
    
    def _handleDataChannel(self):
        @self._dataChannel.on('open')
        def on_open():
            if self._closed:
                return
            logging.debug('DataConsumer DataChannel "open" event')
            self.emit('open')

        # NOTE: aiortc.RTCDataChannel won't emit error event, here use pyee error event
        @self._dataChannel.on('error')
        def on_error(message):
            if self._closed:
                return

            logging.error(f'DataConsumer DataChannel "error" event: {message}')
            
            self.emit('error', message)

        @self._dataChannel.on('close')
        def on_close():
            if self._closed:
                return
            logging.warning('DataConsumer DataChannel "close" event')
            self._closed = True
            self.emit('@close')
            self._observer.emit('close')

        @self._dataChannel.on('message')
        def on_message(message):
            if self._closed:
                return
            self.emit('message', message)



import sys
if sys.version_info >= (3, 8):
    from typing import Callable, Literal, List, Optional, Any
else:
    from typing import Callable, List, Optional, Any
    from typing_extensions import Literal

from pydantic import BaseModel
from aiortc import RTCIceServer, MediaStreamTrack
from ..ortc import ExtendedRtpCapabilities
from ..emitter import EnhancedEventEmitter
from ..models.transport import IceCandidate, IceParameters, DtlsParameters
from ..models.handler_interface import HandlerReceiveOptions, HandlerSendResult, HandlerReceiveResult, SctpStreamParameters, HandlerSendDataChannelResult, HandlerReceiveDataChannelOptions, HandlerReceiveDataChannelResult
from ..rtp_parameters import RtpParameters, RtpCapabilities, RtpCodecCapability, MediaKind, RtpEncodingParameters
from ..sctp_parameters import SctpCapabilities, SctpStreamParameters, SctpParameters
from ..producer import ProducerCodecOptions


class HandlerInterface(EnhancedEventEmitter):
    # @emits @connect - (
	#     { dtlsParameters: DtlsParameters },
	#     callback: Function,
	#     errback: Function
	#   )
	# @emits @connectionstatechange - (connectionState: ConnectionState)
    def __init__(self, loop=None):
        super(HandlerInterface, self).__init__(loop=loop)

    @property
    def name(self) -> str:
        pass
    
    def close(self):
        pass

    async def getNativeRtpCapabilities(self) -> RtpCapabilities:
        pass

    async def getNativeSctpCapabilities(self) -> SctpCapabilities:
        pass

    def run(
        self,
        direction: Literal['send', 'recv'],
        iceParameters: IceParameters,
        iceCandidates: List[IceCandidate],
        dtlsParameters: DtlsParameters,
        extendedRtpCapabilities: ExtendedRtpCapabilities,
        sctpParameters: Optional[SctpParameters]=None,
        iceServers: Optional[RTCIceServer]=None,
        iceTransportPolicy: Optional[Literal['all', 'relay']]=None,
        additionalSettings: Optional[Any]=None,
        proprietaryConstraints: Optional[Any]=None
    ):
        pass

    async def updateIceServers(self, iceServers: List[RTCIceServer]):
        pass

    async def restartIce(self, iceParameters: IceParameters):
        pass

    async def getTransportStats(self) -> Any:
        pass
    
    async def send(
        self,
        track: MediaStreamTrack,
        encodings: List[RtpEncodingParameters]=[],
        codecOptions: Optional[ProducerCodecOptions]=None,
        codec: Optional[RtpCodecCapability]=None
    ) -> HandlerSendResult:
        pass
    
    async def stopSending(self, localId: str):
        pass

    async def replaceTrack(self, localId: str, track: Optional[MediaStreamTrack] = None):
        pass

    async def setMaxSpatialLayer(self, localId: str, spatialLayer: int):
        pass

    async def setRtpEncodingParameters(self, localId: str, params: Any):
        pass

    async def getSenderStats(self, localId: str) -> Any:
        pass

    async def sendDataChannel(
        self,
        streamId: Optional[int]=None,
        ordered: Optional[bool]=True,
        maxPacketLifeTime: Optional[int]=None,
        maxRetransmits: Optional[int]=None,
        priority: Optional[Literal['very-low','low','medium','high']]=None,
        label: Optional[str]=None,
        protocol: Optional[str]=None
    ) -> HandlerSendDataChannelResult:
        pass

    async def receive(
        self,
        trackId: str,
        kind: MediaKind,
        rtpParameters: RtpParameters
    ) -> HandlerReceiveResult:
        pass

    async def stopReceiving(self, localId: str):
        pass

    async def getReceiverStats(self, localId: str):
        pass

    async def receiveDataChannel(
        self,
        sctpStreamParameters: SctpStreamParameters,
        label: Optional[str]=None,
        protocol: Optional[str]=None
    ) -> HandlerReceiveDataChannelResult:
        pass


from .parser import parse, parseParams, parseImageAttributes, parseSimulcastStreamList
from .writer import write, defaultOuterOrder, defaultInnerOrder


class sdp_transform:

    @staticmethod
    def parse(sdp: str) -> dict:
        return parse(sdp)

    @staticmethod
    def parseParams(string: str):
        return parseParams(string)

    @staticmethod
    def parseImageAttributes(string: str):
        return parseImageAttributes(string)

    @staticmethod
    def parseSimulcastStreamList(string: str):
        return parseSimulcastStreamList(string)

    @staticmethod
    def write(session: dict, outerOrder: list = defaultOuterOrder, innerOrder: list = defaultInnerOrder):
        return write(session, outerOrder, innerOrder)



from typing import List, Optional

from ...deps.sdp_transform import sdp_transform
from ...models.transport import DtlsParameters, DtlsRole, DtlsFingerprint
from ...rtp_parameters import RtpCapabilities, RtpCodecCapability, RtpHeaderExtension, RtpParameters, RtcpFeedback


def extractRtpCapabilities(sdpDict: dict) -> RtpCapabilities:
    # Map of RtpCodecParameters indexed by payload type.
    codecsMap:dict = {}
    # Array of RtpHeaderExtensions.
    headerExtensions: List[RtpHeaderExtension] = []
    # Whether a m=audio/video section has been already found.
    gotAudio = False
    gotVideo = False

    m: dict
    for m in sdpDict.get('media', []):
        kind: str = m.get('type')
        if kind == 'audio':
            if gotAudio:
                continue
            gotAudio = True
        elif kind == 'video':
            if gotVideo:
                continue
            getVideo = True
        # Get codecs.
        rtp: dict
        for rtp in m.get('rtp', []):
            codec: RtpCodecCapability = RtpCodecCapability(
                kind=kind,
                mimeType=f"{kind}/{rtp.get('codec','')}",
                preferredPayloadType=rtp.get('payload'),
                clockRate=rtp.get('rate'),
                channels=rtp.get('encoding'),
                parameters={},
                rtcpFeedback=[]
            )
            codecsMap[codec.preferredPayloadType] = codec
        
        # Get codec parameters.
        for fmtp in m.get('fmtp', []):
            parameters = sdp_transform.parseParams(fmtp.get('config'))
            codec = codecsMap.get(fmtp.get('payload'))

            if not codec:
                continue
            # Specials case to convert parameter value to string.
            if parameters.get('profile-level-id'):
                parameters['profile-level-id'] = str(parameters['profile-level-id'])
            codec.parameters = parameters

        # Get RTCP feedback for each codec.
        for fb in m.get('rtcpFb', []):
            codec = codecsMap.get(fb.get('payload'))
            if not codec:
                continue 
            feedback = RtcpFeedback(
                type=fb.get('type'),
                parameter=fb.get('subtype', '')
            )
            codec.rtcpFeedback.append(feedback)

        # Get RTP header extensions.
        for ext in m.get('ext'):
            if ext.get('encrypt-uri'):
                continue
            headerExtension: RtpHeaderExtension = RtpHeaderExtension(
                kind=kind,
                uri=ext.get('uri'),
                preferredId=ext.get('value')
            )
            headerExtensions.append(headerExtension)

    rtpCapabilities: RtpCapabilities = RtpCapabilities(
        codecs=list(codecsMap.values()),
        headerExtensions=headerExtensions
    )

    return rtpCapabilities

def extractDtlsParameters(sdpDict: dict) -> DtlsParameters:
    mediaDicts = [m for m in sdpDict.get('media', []) if m.get('iceUfrag') and m.get('port') != 0]
    if not mediaDicts:
        raise Exception('no active media section found')
    mediaDict = mediaDicts[0]
    fingerprint = mediaDict.get('fingerprint') if mediaDict.get('fingerprint') else sdpDict.get('fingerprint')
    role = 'auto'
    if mediaDict.get('setup') == 'activate':
        role = 'client'
    elif mediaDict.get('setup') == 'passive':
        role = 'server'
    elif mediaDict.get('setup') == 'actpass':
        role = 'auto'
    
    dtlsParameters = DtlsParameters(
        role=role,
        fingerprints=[DtlsFingerprint(
            algorithm=fingerprint.get('type'),
            value=fingerprint.get('hash')
        )]
    )

    return dtlsParameters

def getCname(offerMediaDict: dict):
    ssrcCnameLines = [line for line in offerMediaDict.get('ssrcs', []) if line.get('attribute') == 'cname']
    if not ssrcCnameLines:
        return ''
    return ssrcCnameLines[0].get('value', '')

def applyCodecParameters(offerRtpParameters: RtpParameters, answerMediaDict: Optional[dict] = {}):
    for codec in offerRtpParameters.codecs:
        mimeType = codec.mimeType.lower()
        if mimeType != 'audio/opus':
            continue

        rtps = [r for r in answerMediaDict.get('rtp', []) if r.get('payload') == codec.payloadType]
        if not rtps:
            continue
        rtp = rtps[0]
        
        fmtps = [f for f in answerMediaDict.get('rmtp', []) if f.get('payload') == codec.payloadType]
        if not fmtps:
            fmtp = {
                'payload': codec.payloadType,
                'config': ''
            }
            if answerMediaDict.get('rmtp') != None:
                answerMediaDict['fmtp'].append(fmtp)
            else:
                answerMediaDict['fmtp'] = [fmtp]
        else:
            fmtp = fmtps[0]
        
        parameters = sdp_transform.parseParams(fmtp.get('config', ''))
        
        if mimeType == 'audio/opus':
            spropStereo = codec.parameters.get('sprop-stereo')
            if spropStereo != None:
                parameters['stereo'] = 1 if spropStereo else 0

        # Write the codec fmtp.config back.
        fmtp['config'] = ';'.join([f'{key}={value}' for key, value in parameters.items()])



import asyncio
import logging

from smcdk.log import Logger
from smcdk.api.mediasoup_signaler import MediasoupSignalerInterface, ProtooSignaler, MessageType, Request
from smcdk.api.multimedia_runtime import MultimediaRuntime
from smcdk.api.notification_listener import BandwidthNotificationListener, PeerNotificationListener, \
    ProducerNotificationListener, ConsumerNotificationListener, DataConsumerNotificationListener
from smcdk.api.request_listener import ConsumerRequestListener, DataConsumerRequestListener
from smcdk.api.room_peer import Room, Peer, PeerAppData

# logger of module level
logger = Logger.getLogger(enable_console=True, level=logging.WARN, log_file_path=None)


class MediasoupClient:

    def __init__(self,
                 signaler: MediasoupSignalerInterface = None,
                 requestListeners: list = None,
                 notificationListeners: list = None):
        """
        instantiate a MediasoupClient object

        :param signaler:
            pluggable signaler implementation of MediasoupSignalerInterface,
            if not provided, default is protooSignaler
        :param requestListeners:
            request listener list of consumer and dataConsumer level,
            if not provided, default is [ConsumerRequestListener, DataConsumerRequestListener],
            if provided, inherit and overwrite these default request listeners
        :param notificationListeners:
            notification listener list,
            if not provided, default is
                [BandwidthNotificationListener, PeerNotificationListener, ProducerNotificationListener,
                    ConsumerNotificationListener, DataConsumerNotificationListener],
            if provided, inherit and overwrite these default notification listeners
        """
        '''
        multimedia runtime, room and peer part        
        '''
        self._multimediaRuntime: MultimediaRuntime = MultimediaRuntime()
        # room should be initialized to be injected into peer
        self._room: Room = Room()
        # peer should be initialized to be injected into listeners
        self._mePeer: Peer = Peer(room=self._room)
        '''
        signaler part
        '''
        if signaler is None:
            self._signaler: MediasoupSignalerInterface = ProtooSignaler()
        else:
            self._signaler: MediasoupSignalerInterface = signaler
        '''
        loop and tasks part
        '''
        self._loop = None
        self._loopTasks = []
        '''
        request listeners
        '''
        if requestListeners is None:
            self._consumerRequestListener: ConsumerRequestListener = ConsumerRequestListener(self._mePeer)
            self._dataConsumerRequestListener: DataConsumerRequestListener = DataConsumerRequestListener(self._mePeer)
        elif len(list) == 2:
            self._consumerRequestListener: ConsumerRequestListener = requestListeners[0]
            self._consumerRequestListener.mePeer = self._mePeer
            self._dataConsumerRequestListener: DataConsumerRequestListener = requestListeners[1]
            self._dataConsumerRequestListener.mePeer = self._mePeer
        else:
            raise Exception(f"wrong length of requestListenerList: {len(requestListeners)}, expect: 2")
        '''
        notification listeners
        '''
        if notificationListeners is None:
            self._bandwidthNotificationListener: BandwidthNotificationListener = BandwidthNotificationListener(
                self._mePeer)
            self._peerNotificationListener: PeerNotificationListener = PeerNotificationListener(self._mePeer)
            self._producerNotificationListener: ProducerNotificationListener = ProducerNotificationListener(
                self._mePeer)
            self._consumerNotificationListener: ConsumerNotificationListener = ConsumerNotificationListener(
                self._mePeer)
            self._dataConsumerNotificationListener: DataConsumerNotificationListener = DataConsumerNotificationListener(
                self._mePeer)
        elif len(notificationListeners) == 5:
            self._bandwidthNotificationListener: BandwidthNotificationListener = notificationListeners[0]
            self._bandwidthNotificationListener.mePeer = self._mePeer
            self._peerNotificationListener: PeerNotificationListener = notificationListeners[1]
            self._peerNotificationListener.mePeer = self._mePeer
            self._producerNotificationListener: ProducerNotificationListener = notificationListeners[2]
            self._producerNotificationListener.mePeer = self._mePeer
            self._consumerNotificationListener: ConsumerNotificationListener = notificationListeners[3]
            self._consumerNotificationListener.mePeer = self._mePeer
            self._dataConsumerNotificationListener: DataConsumerNotificationListener = notificationListeners[4]
            self._dataConsumerNotificationListener.mePeer = self._mePeer
        else:
            raise Exception(f"wrong length of notificationListeners: {len(notificationListeners)}, expect: 5")
        self._notificationListeners = [self._bandwidthNotificationListener, self._peerNotificationListener,
                                       self._producerNotificationListener, self._consumerNotificationListener,
                                       self._dataConsumerNotificationListener]

    # async def joinSingleRoom(self, roomAddressInfo: dict, peerInfo: dict,
    #                          producerConfig: dict,
    #                          consumerConfig: dict):
    #     try:
    #         await self.joinRoom(roomAddressInfo, peerInfo, producerConfig, consumerConfig)
    #     except (asyncio.CancelledError, KeyboardInterrupt):
    #         await self.close()

    async def joinRoom(self, roomAddressInfo: dict, peerInfo: dict, producerConfig: dict, consumerConfig: dict):
        """
        enter the specified room and auto-produce if needed

        :param roomAddressInfo:
            {
                'serverAddress':
                    str, ip:port, required,
                'enableSslVerification':
                    bool, will be sent to MediasoupSignalerInterface.connectToRoom‘s 'enableSslVerification' argument,
                    to check if ssl verification is needed before establish connection to server,
                    the default value is True, that means ssl verification is needed
                'roomId': 'the room's id, required',
            }
        :param peerInfo:
            {
                'peerId':
                    str, the peer id when current MediasoupClient instance join the room, required
                'displayName':
                    str, the peer id when current MediasoupClient instance join the room, default is 'PySDK Client'
                'device':
                    dict, the device info when current MediasoupClient instance join the room,
                    default is
                        {
                            'flag': 'smcdk,pymediasoup,aiortc,python',
                            'name': 'mediasoup-client-pysdk',
                            'version': '1.x'
                        }
            }
        :param producerConfig:
            {
                'autoProduce': bool, default is True
                'mediaFilePath': str, the full path of media file, required
            }
        :param consumerConfig:
            {
                'autoConsume': bool, default is True
                'recordDirectoryPath': str, the root path of to-record media files, required
            }
        :return: None
        """
        logger.info('peer(id=%s, displayName=%s) join room(id=%s)', peerInfo['peerId'], peerInfo['displayName'],
                    roomAddressInfo['roomId'])

        '''
        stuff room and peer part
        '''
        # not first enter
        if self._room.roomId is not None:
            logger.warn('exit from room(id=%s) before join room(id=%s)', self._room.roomId, roomAddressInfo['roomId'])
            await self.exitRoom(self._room.roomId)
        self._room.serverAddress = roomAddressInfo['serverAddress']
        self._room.roomId = roomAddressInfo['roomId']
        self._mePeer.peerId = peerInfo['peerId']
        self._mePeer.data = PeerAppData(displayName=peerInfo.get('displayName', 'PySDK Client'),
                                        device=peerInfo.get('device', {'flag': 'smcdk,pymediasoup,aiortc,python',
                                                                       'name': 'mediasoup-client-pysdk',
                                                                       'version': '1.x'}))
        '''
        stuff multimedia runtime part
        '''
        self._multimediaRuntime.initializeProducerAndConsumerOptions(
            autoProduce=producerConfig.get('autoProduce', True),
            mediaFilePath=producerConfig.get('mediaFilePath'),
            autoConsume=consumerConfig.get('autoConsume', True),
            recordDirectoryPath=consumerConfig.get(
                'recordDirectoryPath'),
            recordFilePathGenerator=consumerConfig.get('recordFilePathGenerator')
        )
        '''
        create connection to server by signaler
        '''
        self._loop = asyncio.get_running_loop()
        logger.info('connectToRoom, serverAddress=%s, roomId=%s, peerId=%s', self._room.serverAddress,
                    self._room.roomId,
                    self._mePeer.peerId)
        await self._signaler.connectToRoom(self._loop, self._room.serverAddress, self._room.roomId, self._mePeer.peerId,
                                           roomAddressInfo['enableSslVerification'])
        '''
        create loop tasks
        '''
        bandwidthNotificationLoop = self._loop.create_task(
            self._bandwidthNotificationListener.runLoop(asyncio.Queue(loop=self._loop)))
        peerNotificationLoop = self._loop.create_task(
            self._peerNotificationListener.runLoop(asyncio.Queue(loop=self._loop)))
        producerNotificationLoop = self._loop.create_task(
            self._producerNotificationListener.runLoop(asyncio.Queue(loop=self._loop)))
        consumerNotificationLoop = self._loop.create_task(
            self._consumerNotificationListener.runLoop(asyncio.Queue(loop=self._loop)))
        dataConsumerNotificationLoop = self._loop.create_task(
            self._dataConsumerNotificationListener.runLoop(asyncio.Queue(loop=self._loop)))
        severEventLoop = self._loop.create_task(self._serverEventLoop())
        self._loopTasks = [severEventLoop, bandwidthNotificationLoop, peerNotificationLoop, producerNotificationLoop,
                           consumerNotificationLoop, dataConsumerNotificationLoop]
        '''
        load device
        '''
        await self._loadDeviceByRouterRtpCapabilities()
        if not (self._multimediaRuntime.canProduce or self._multimediaRuntime.canConsume):
            return
        '''
        create sendTransport & recvTransport 
        '''
        if self._multimediaRuntime.canProduce:
            await self._createSendTransport()
        if self._multimediaRuntime.canConsume:
            await self._createRecvTransport()
        '''
        formally join
        '''
        await self._joinFormally()
        '''
        produce(push media stream to the server) automatically if needed 
        '''
        if self._multimediaRuntime.autoProduce:
            # self._loop.create_task(self._produce(kind='audio', source='mic'), name='produceTask_audio_mic')
            # self._loop.create_task(self._produce(kind='video', source='webCam'), name='produceTask_video_webCam')
            # self._loop.create_task(self._produce(kind='video', source='screenShare'),
            #                        name='produceTask_video_screenShare')
            self._loop.create_task(self._produce())
        '''
        waiting loop tasks
        '''
        await asyncio.gather(severEventLoop, bandwidthNotificationLoop, peerNotificationLoop, producerNotificationLoop,
                             consumerNotificationLoop, dataConsumerNotificationLoop, return_exceptions=True)

    def play(self):
        if self._loop is not None:
            return self._loop.create_task(self._produce())
        else:
            logger.error(f'can\'t play, missing loop({self._loop})')

    async def exitRoom(self, roomId):
        """
        exit current room, prepare to join another
        :param roomId: current room's id to exit, required
        :return: None
        """

        def stopTaskFunc():
            for task in self._loopTasks:
                task.cancel()
            # quick GC, not needed currently
            # for notificationListeners in self._notificationListeners:
            #     notificationListeners.resetQueue(asyncio.Queue)

        await self._multimediaRuntime.close(stopTaskFunc)
        self._room.serverAddress = None
        self._room.roomId = None
        logger.info('exit room %s finished', roomId)

    async def close(self):
        """
        stop current MediasoupClient forever, for purpose of exiting process/thread only

        :return: None
        """
        if self._room.roomId is not None:
            await self.exitRoom(roomId=self._room.roomId)
        else:
            logger.warn('already closed')

    async def _loadDeviceByRouterRtpCapabilities(self):
        requestId = await self._signaler.getRouterRtpCapabilities()
        logger.info('signal request: getRouterRtpCapabilities, requestId=%s', requestId)
        response = await self._signaler.getResponse(requestId)
        await self._multimediaRuntime.loadDevice(response.data)

    async def _createSendTransport(self):
        if self._multimediaRuntime.sendTransportId is not None:
            logger.warn('the sendTransport has already been created')
            return
        requestId = await self._signaler.createSendTransport(self._multimediaRuntime.sctpCapabilities.dict())
        logger.info('signal request: createWebRtcTransport, direction=send, requestId=%s', requestId)
        response = await self._signaler.getResponse(requestId)

        async def onConnect(dtlsParameters):
            requestIdToConnectSWT = await self._signaler.connectWebRtcTransport(self._multimediaRuntime.sendTransportId,
                                                                                dtlsParameters.dict(exclude_none=True))
            logger.info('signal request: connectWebRtcTransport, direction: send, requestId: %s', requestIdToConnectSWT)
            await self._signaler.getResponse(requestIdToConnectSWT)

        async def onProduce(kind: str, rtpParameters, appData: dict):
            requestIdToProduce = await self._signaler.produce(self._multimediaRuntime.sendTransportId, kind,
                                                              rtpParameters.dict(exclude_none=True), appData)
            logger.info('signal request: produce, requestId=%s', requestIdToProduce)
            response_ = await self._signaler.getResponse(requestIdToProduce)
            return response_.data['id']

        async def onProduceData(
                sctpStreamParameters,
                label: str,
                protocol: str,
                appData: dict
        ):
            requestIdToProduceData = await self._signaler.produceData(self._multimediaRuntime.sendTransportId, label,
                                                                      protocol,
                                                                      sctpStreamParameters.dict(exclude_none=True),
                                                                      appData)
            logger.info('signal request: produceData, requestId=%s', requestIdToProduceData)
            response_ = await self._signaler.getResponse(requestIdToProduceData)
            return response_.data['id']

        self._multimediaRuntime.createSendTransport(transportId=response.data['id'],
                                                    iceParameters=response.data['iceParameters'],
                                                    iceCandidates=response.data['iceCandidates'],
                                                    dtlsParameters=response.data['dtlsParameters'],
                                                    sctpParameters=response.data['sctpParameters'],
                                                    onConnectFunc=onConnect,
                                                    onProduceFunc=onProduce,
                                                    onProduceDataFunc=onProduceData)

    async def _createRecvTransport(self):
        if self._multimediaRuntime.recvTransportId is not None:
            logger.warn('the recvTransport has already been created')
            return
        requestId = await self._signaler.createRecvTransport(self._multimediaRuntime.sctpCapabilities.dict())
        logger.info('signal request: createWebRtcTransport, direction=recv, requestId=%s', requestId)
        response = await self._signaler.getResponse(requestId)

        async def onConnect(dtlsParameters):
            requestIdToConnectRWT = await self._signaler.connectWebRtcTransport(self._multimediaRuntime.recvTransportId,
                                                                                dtlsParameters.dict(exclude_none=True))
            logger.info('signal request: connectWebRtcTransport, direction: recv, requestId=%s', requestIdToConnectRWT)
            await self._signaler.getResponse(requestIdToConnectRWT)

        self._multimediaRuntime.createRecvTransport(transportId=response.data['id'],
                                                    iceParameters=response.data['iceParameters'],
                                                    iceCandidates=response.data['iceCandidates'],
                                                    dtlsParameters=response.data['dtlsParameters'],
                                                    sctpParameters=response.data['sctpParameters'],
                                                    onConnectFunc=onConnect)

    async def _joinFormally(self):
        requestId = await self._signaler.join(self._mePeer.data.displayName,
                                              self._mePeer.data.device,
                                              self._multimediaRuntime.rtpCapabilities.dict(exclude_none=True),
                                              self._multimediaRuntime.sctpCapabilities.dict(exclude_none=True))
        logger.info('signal request: join, requestId=%s', requestId)
        message = await self._signaler.getResponse(requestId)
        for peerInfo in message.data['peers']:
            self._room.addPeer(peerInfo['id'], data=PeerAppData(displayName=peerInfo['displayName'],
                                                                device=peerInfo['device']))

    # enableShare
    async def _produce(self):
        if not self._multimediaRuntime.canProduce:
            logger.warn('can\'t play, please check mediaFilePath and rtpCapabilities')
            return
        if self._multimediaRuntime.sendTransportId is None:
            await self._createSendTransport()
        await self._multimediaRuntime.produce()

        # record

    async def _consume(self, message: Request):
        if not self._multimediaRuntime.canConsume or not self._multimediaRuntime.autoConsume:
            return
        if self._multimediaRuntime.recvTransportId is None:
            await self._createRecvTransport()
        consumerId = message.data['id']
        peerId = message.data['peerId']
        producerPeer = self._mePeer.room.getPeerByPeerId(peerId)
        producerId = message.data['producerId']
        kind = message.data['kind']
        rtpParameters = message.data['rtpParameters']
        await self._multimediaRuntime.consume(self._mePeer, consumerId, producerPeer, producerId, kind, rtpParameters)
        otherPeer = self._room.getPeerByPeerId(peerId)
        self._room.bindConsumerIdToPeer(consumerId, otherPeer)
        requestId = message.requestId
        await self._consumerRequestListener.onNewConsumer(self._signaler.responseToNewConsumer(requestId), message,
                                                          otherPeer)

    async def _consumeData(self, message: Request):
        if self._multimediaRuntime.recvTransportId is None:
            await self._createRecvTransport()
        dataConsumerId = message.data['id']
        dataProducerId = message.data['dataProducerId']
        sctpStreamParameters = message.data['sctpStreamParameters']
        label = message.data['label']
        protocol = message.data['protocol']
        appData = message.data['appData']

        def onMessage(recvMessage):
            logger.info('DataChannel {%s}-{%s}: {%s}', label, protocol, recvMessage)
            self._dataConsumerNotificationListener.onMessage(otherPeer, recvMessage, label, protocol, appData)

        await self._multimediaRuntime.consumeData(dataConsumerId, dataProducerId, sctpStreamParameters, label, protocol,
                                                  appData, onMessage)
        otherPeer = self._room.getPeerByPeerId(message.data['peerId'])
        self._room.bindDataConsumerIdToPeer(dataConsumerId, otherPeer)
        requestId = message.requestId
        await self._dataConsumerRequestListener.onNewDataConsumer(self._signaler.responseToNewDataConsumer(requestId),
                                                                  message,
                                                                  otherPeer)

    async def _serverEventLoop(self):
        while True:
            message = await self._signaler.receiveMessage()
            if message.get('response'):
                logger.info('receive response for requestId: %d', message['id'])
                logger.debug('response details: ok=%s, data=%s', message['data'])
                self._signaler.setResponse(message)
            elif message.get('request'):
                logger.info('receive request, requestId=%d, method=%s', message['id'], message['method'])
                if message['method'] == MessageType.SERVER_REQURST_newConsumer.value:
                    # don't use: asyncio.create_task(self._consume(...))
                    # to prevent that consumer's notification precede it‘s request
                    await self._consume(Request(message['id'], message['method'], message['data']))
                elif message['method'] == MessageType.SERVER_REQURST_newDataConsumer.value:
                    await self._consumeData(Request(message['id'], message['method'], message['data']))
                else:
                    logger.error('unhandled request: %s' + message)
            elif message.get('notification'):
                logger.info('receive notification, method=%s', message['method'])
                if message['method'] in {MessageType.SERVER_NOTIFICATION_downlinkBwe.value}:
                    await self._bandwidthNotificationListener.enqueue(message)
                elif message['method'] in {MessageType.SERVER_NOTIFICATION_activeSpeaker.value,
                                           MessageType.SERVER_NOTIFICATION_newPeer.value,
                                           MessageType.SERVER_NOTIFICATION_peerDisplayNameChanged.value,
                                           MessageType.SERVER_NOTIFICATION_peerClosed.value}:
                    await self._peerNotificationListener.enqueue(message)
                elif message['method'] in {MessageType.SERVER_NOTIFICATION_producerScore.value}:
                    await self._producerNotificationListener.enqueue(message)
                elif message['method'] in {MessageType.SERVER_NOTIFICATION_consumerScore.value,
                                           MessageType.SERVER_NOTIFICATION_consumerLayersChanged.value,
                                           MessageType.SERVER_NOTIFICATION_consumerPaused.value,
                                           MessageType.SERVER_NOTIFICATION_consumerResumed.value,
                                           MessageType.SERVER_NOTIFICATION_consumerClosed.value}:
                    await self._consumerNotificationListener.enqueue(message)
                elif message['method'] in {MessageType.SERVER_NOTIFICATION_dataConsumerClosed.value}:
                    await self._dataConsumerNotificationListener.enqueue(message)
                else:
                    logger.error('unhandled notification: %s' + message)
            # bypass other no-exists message type



import logging
from typing import Optional, Any
from aiortc import RTCRtpReceiver, MediaStreamTrack
from pyee import AsyncIOEventEmitter
from pydantic import BaseModel
from .errors import InvalidStateError, UnsupportedError
from .emitter import EnhancedEventEmitter
from .rtp_parameters import MediaKind, RtpParameters


class ConsumerOptions(BaseModel):
    id: str
    producerId: str
    kind: MediaKind
    rtpParameters: RtpParameters
    appData: Optional[dict] = {}

class Consumer(EnhancedEventEmitter):
    def __init__(
        self,
        id: str,
        localId: str,
        producerId: str,
        track: MediaStreamTrack,
        rtpParameters: RtpParameters,
        rtpReceiver: Optional[RTCRtpReceiver] = None,
        appData: Optional[dict] = {},
        loop=None
    ):
        super(Consumer, self).__init__(loop=loop)

        # Closed flag.
        self._closed: bool = False
        # Observer instance.
        self._observer: AsyncIOEventEmitter = AsyncIOEventEmitter()

        self._id = id
        self._localId = localId
        self._producerId = producerId
        self._track = track
        # NOTE: 'AudioStreamTrack' object has no attribute 'enabled'
        self._paused: bool = False
        self._rtpParameters = rtpParameters
        self._rtpReceiver = rtpReceiver
        self._appData = appData
        
        self._handleTrack()
    
    # Consumer id.
    @property
    def id(self) -> str:
        return self._id
    
    # Local id.
    @property
    def localId(self) -> str:
        return self._localId

    # Associated Producer id.
    @property
    def producerId(self) -> str:
        return self._producerId
    
    # Whether the Consumer is closed.
    @property
    def closed(self) -> bool:
        return self._closed
    
    # Media kind.
    @property
    def kind(self) -> MediaStreamTrack.kind:
        return self._track.kind
    
    # Associated RTCRtpReceiver.
    @property
    def rtpReceiver(self) -> Optional[RTCRtpReceiver]:
        return self._rtpReceiver
    
    # The associated track.
    @property
    def track(self) -> Optional[MediaStreamTrack]:
        return self._track
    
    # RTP parameters.
    @property
    def rtpParameters(self) -> RtpParameters:
        return self._rtpParameters
    
    # Whether the Consumer is paused.
    @property
    def paused(self) -> bool:
        return self._paused
    
    # App custom data.
    @property
    def appData(self) -> Any:
        return self._appData
    
    # Invalid setter.
    @appData.setter
    def appData(self, value):
        raise Exception('cannot override appData object')

    
    # Observer.
    #
    # @emits close
    # @emits pause
    # @emits resume
    # @emits trackended
    @property
    def observer(self) -> AsyncIOEventEmitter:
        return self._observer

    async def close(self):
        if self._closed:
            return
        
        logging.debug('Consumer close()')

        self._closed = True

        self._destroyTrack()

        await self.emit_for_results('@close')

        # Emit observer event.
        self._observer.emit('close')
  
    # Transport was closed.
    def transportClosed(self):
        if self._closed:
            return

        logging.debug('Consumer transportClosed()')

        self._closed = True

        self._destroyTrack()

        self.emit('transportclose')

        self._observer.emit('close')
    
    # Get associated RTCRtpSender stats.
    async def getStats(self):
        if self._closed:
            raise InvalidStateError('closed')

        return await self.emit_for_results('@getstats')
    
    # Pauses sending media.
    def pause(self):
        logging.debug('Consumer pause()')

        if self._closed:
            logging.debug('Consumer pause() | Consumer closed')
            return
        
        self._paused = True

        if self._track and self._disableTrackOnPause:
            # TODO: MediaStreamTrack missing enable property
            # self._track.enabled = False
            pass
        
        self._observer.emit('pause')
    
    # Resumes sending media.
    def resume(self):
        logging.debug('Consumer resume()')

        if self._closed:
            logging.debug('Consumer resume() | Consumer closed')
            return
        
        self._paused = False

        if self._track and self._disableTrackOnPause:
            # TODO: MediaStreamTrack missing enable property
            # self._track.enabled = True
            pass
        
        self._observer.emit('resume')
    
    def _onTrackEnded(self):
        logging.debug('track "ended" event')
        self.emit('trackended')
        # Emit observer event.
        self._observer.emit('trackended')
    
    def _handleTrack(self):
        if not self._track:
            return

        self._track.on('ended', self._onTrackEnded)
    
    def _destroyTrack(self):
        if not self._track:
            return

        self._track.remove_listener('ended', self._onTrackEnded)

        self._track.stop()



import asyncio
import json
# for ProtooSignaler
import random
import ssl
from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import Dict

import websockets


class MediasoupSignalerInterface(metaclass=ABCMeta):

    @abstractmethod
    async def connectToRoom(self, loop, serverAddress, roomId, peerId, enableSslVerification: bool = True):
        pass

    @abstractmethod
    async def getRouterRtpCapabilities(self) -> int:
        pass

    @abstractmethod
    async def createSendTransport(self, sctpCapabilities: dict):
        pass

    @abstractmethod
    async def createRecvTransport(self, sctpCapabilities: dict):
        pass

    @abstractmethod
    async def join(self, displayName: str, device: dict, rtpCapabilities: dict, sctpCapabilities: dict):
        pass

    @abstractmethod
    async def connectWebRtcTransport(self, transportId: str, dtlsParameters: dict):
        pass

    @abstractmethod
    async def produce(self, transportId: str, kind: str, rtpParameters: dict, appData: dict):
        pass

    @abstractmethod
    async def produceData(self, transportId: str, label: str, protocol: str, sctpStreamParameters: dict, appData: dict):
        pass

    @abstractmethod
    async def receiveMessage(self):
        pass

    @abstractmethod
    async def responseToNewConsumer(self, requestId: int):
        pass

    @abstractmethod
    async def responseToNewDataConsumer(self, requestId: int):
        pass

    @abstractmethod
    def setResponse(self, message: dict):
        pass

    @abstractmethod
    async def getResponse(self, requestId: int):
        pass


class ProtooSignaler(MediasoupSignalerInterface):

    def __init__(self):
        self._loop = None
        self._ctx = None
        self._websocket = None
        self._roomUri = None
        self._responses: Dict[int, asyncio.Future] = {}

    async def connectToRoom(self, loop: asyncio.AbstractEventLoop, serverAddress, roomId,
                            peerId, enableSslVerification: bool = True):  # todo: 格式化，并存储roomId
        self._loop = loop
        self._roomUri = f'wss://{serverAddress}/?roomId={roomId}&peerId={peerId}'
        self._ctx = ssl.create_default_context()
        if not enableSslVerification:
            self._ctx.check_hostname = False
            self._ctx.verify_mode = ssl.CERT_NONE
        self._websocket = await websockets.connect(self._roomUri, subprotocols=['protoo'], ssl=self._ctx)

    async def closeCurrentConnection(self):
        self._websocket.close()
        self._loop = None

    @staticmethod
    def generateRandomNumber() -> int:
        return round(random.random() * 10000000)

    async def getRouterRtpCapabilities(self) -> int:
        return await self._send_request({
            'request': True,
            'method': 'getRouterRtpCapabilities',
            'data': {}
        })

    async def createSendTransport(self, sctpCapabilities: dict):
        return await self._send_request({
            'request': True,
            'method': 'createWebRtcTransport',
            'data': {
                'forceTcp': False,
                'producing': True,
                'consuming': False,
                'sctpCapabilities': sctpCapabilities
            }
        })

    async def createRecvTransport(self, sctpCapabilities: dict):
        return await self._send_request({
            'request': True,
            'method': 'createWebRtcTransport',
            'data': {
                'forceTcp': False,
                'producing': False,
                'consuming': True,
                'sctpCapabilities': sctpCapabilities
            }
        })

    async def join(self, displayName: str, device: dict, rtpCapabilities: dict, sctpCapabilities: dict):
        return await self._send_request({
            'request': True,
            'method': 'join',
            'data': {
                "displayName": displayName,
                'device': device,
                "rtpCapabilities": rtpCapabilities,
                "sctpCapabilities": sctpCapabilities
            }
        })

    async def connectWebRtcTransport(self, transportId: str, dtlsParameters: dict):
        return await self._send_request({
            "request": True,
            "method": "connectWebRtcTransport",
            "data": {
                "transportId": transportId,
                "dtlsParameters": dtlsParameters
            }
        })

    async def produce(self, transportId: str, kind: str, rtpParameters: dict, appData: dict):
        return await self._send_request({
            'method': 'produce',
            'request': True,
            'data': {
                'transportId': transportId,
                'kind': kind,
                'rtpParameters': rtpParameters,
                'appData': appData
            }
        })

    async def produceData(self, transportId: str, label: str, protocol: str, sctpStreamParameters: dict, appData: dict):
        return await self._send_request({
            'method': 'produceData',
            'request': True,
            'data': {
                'transportId': transportId,
                'label': label,
                'protocol': protocol,
                'sctpStreamParameters': sctpStreamParameters,
                'appData': appData
            }
        })

    async def _send_request(self, requestParameters: dict) -> int:
        requestParameters['id'] = ProtooSignaler.generateRandomNumber()
        self._responses[requestParameters['id']] = self._loop.create_future()
        await self._websocket.send(json.dumps(requestParameters))
        return requestParameters['id']

    async def _send_response(self, responseParameters: dict):
        await self._websocket.send(json.dumps(responseParameters))

    async def receiveMessage(self):
        return json.loads(await self._websocket.recv())

    async def responseToNewConsumer(self, requestId: str):
        return await self._send_response({
            'response': True,
            'id': requestId,
            'ok': True,
            'data': {}
        })

    async def responseToNewDataConsumer(self, requestId: str):
        return await self._send_response({
            'response': True,
            'id': requestId,
            'ok': True,
            'data': {}
        })

    def setResponse(self, message: dict):
        self._responses[message['id']].set_result(message)

    async def getResponse(self, requestId: int):
        # print(f'getResponse: {requestId}')
        try:
            message = await asyncio.wait_for(fut=self._responses[requestId], timeout=10)
            return Response(requestId=message['id'], method=None, data=message['data'])
        except asyncio.TimeoutError:
            raise Exception("operation timed out")


class MessageType(Enum):
    CLIENT_REQUEST_getRouterRtpCapabilities = 'getRouterRtpCapabilities'
    SERVER_REQURST_newPeer = 'newPeer'
    SERVER_REQURST_newConsumer = 'newConsumer'
    SERVER_REQURST_newDataConsumer = 'newDataConsumer'
    # bandwidth
    SERVER_NOTIFICATION_downlinkBwe = 'downlinkBwe'
    # peer
    SERVER_NOTIFICATION_activeSpeaker = 'activeSpeaker'
    SERVER_NOTIFICATION_newPeer = 'newPeer'
    SERVER_NOTIFICATION_peerDisplayNameChanged = 'peerDisplayNameChanged'
    SERVER_NOTIFICATION_peerClosed = 'peerClosed'
    # producer
    SERVER_NOTIFICATION_producerScore = 'producerScore'
    # consumer
    SERVER_NOTIFICATION_consumerScore = 'consumerScore'
    SERVER_NOTIFICATION_consumerLayersChanged = 'consumerLayersChanged'
    SERVER_NOTIFICATION_consumerPaused = 'consumerPaused'
    SERVER_NOTIFICATION_consumerResumed = 'consumerResumed'
    SERVER_NOTIFICATION_consumerClosed = 'consumerClosed'
    # dataConsumer
    SERVER_NOTIFICATION_dataConsumerClosed = 'dataConsumerClosed'


class SignalerMessage:
    def __init__(self, requestId: int, method: str, data: dict):
        self._requestId = requestId
        self._method = method
        self._data = data

    @property
    def requestId(self) -> int:
        return self._requestId

    @property
    def method(self) -> str:
        return self._method

    @property
    def data(self) -> dict:
        return self._data


class Request(SignalerMessage):
    pass


class Response(SignalerMessage):
    pass


class Notification(SignalerMessage):
    pass



import os
from typing import Union, Optional, Literal

from aiortc import VideoStreamTrack
from aiortc.contrib.media import MediaPlayer, MediaBlackhole, MediaRecorder
from aiortc.mediastreams import AudioStreamTrack

from smcdk.consumer import Consumer
from smcdk.data_consumer import DataConsumer
from smcdk.device import Device
from smcdk.handlers.aiortc_handler import AiortcHandler
from smcdk.producer import Producer
from smcdk.rtp_parameters import RtpCapabilities
from smcdk.sctp_parameters import SctpCapabilities, SctpStreamParameters
from smcdk.transport import Transport
from .room_peer import Peer


class MultimediaRuntime:
    def __init__(self):
        # original mediasoup device
        self._device: Device = None
        '''
        sendTransport and its producers part
        '''
        self._sendTransport: Transport = None
        self._producers: list = []
        self._autoProduce: bool = True
        self._canProduce: bool = False
        self._player: MediaPlayer = None
        self._mediaFilePath: str = None
        self._videoTrack: VideoStreamTrack = None
        self._audioTrack: AudioStreamTrack = None
        self._tracks: list = []
        '''
        recvTransport and its consumers(include dataConsumers) part
        '''
        self._recvTransport: Transport = None
        self._dataConsumers: list = []
        self._consumers: list = []
        self._autoConsume: bool = True
        self._canConsume: bool = True
        self._recordDirectoryPath = None
        # self._recordFilePathGenerator:function
        self._recorders: dict = {}

    @property
    def autoProduce(self) -> bool:
        return self._autoProduce

    @property
    def canProduce(self) -> bool:
        return self._canProduce

    @property
    def autoConsume(self) -> bool:
        return self._autoConsume

    @property
    def canConsume(self) -> bool:
        return self._canConsume

    def initializeProducerAndConsumerOptions(self, autoProduce: bool, mediaFilePath: str, autoConsume: bool,
                                             recordDirectoryPath: str, recordFilePathGenerator):
        """"""
        '''
        producer part
        '''
        self._autoProduce = autoProduce
        self._mediaFilePath = mediaFilePath
        self._canProduce = self._mediaFilePath is not None
        '''        
        consumer part 
        '''
        self._autoConsume = autoConsume
        self._recordDirectoryPath = recordDirectoryPath
        # if recordFilePathGenerator:
        #     self._recordFilePathGenerator = recordFilePathGenerator
        # else:
        #     self._recordFilePathGenerator = self._generateRecordFilePath
        self._canConsume = self._recordDirectoryPath is not None

    def _preparePlayerEngine(self):
        if self._mediaFilePath != '':
            self._player = MediaPlayer(self._mediaFilePath)
        if self._player and self._player.video:
            self._videoTrack = self._player.video
        elif self._mediaFilePath == '':
            self._videoTrack = VideoStreamTrack()
        self._tracks.append(self._videoTrack)
        if self._player and self._player.audio:
            self._audioTrack = self._player.audio
        elif self._mediaFilePath == '':
            self._audioTrack = AudioStreamTrack()
        self._tracks.append(self._audioTrack)

    async def loadDevice(self, routerRtpCapabilities: Union[RtpCapabilities, dict]):
        if len(self._tracks) == 0:
            self._preparePlayerEngine()
        self._device = Device(handlerFactory=AiortcHandler.createFactory(tracks=self._tracks))
        await self._device.load(routerRtpCapabilities)
        self._canProduce &= self._device.canProduce('audio') or self._device.canProduce('video')
        # MediaBlackhole is always able to consume
        # self._canConsume = True

    @property
    def rtpCapabilities(self) -> Optional[RtpCapabilities]:
        if self._device:
            return self._device.rtpCapabilities
        else:
            raise Exception('device has not loaded')

    @property
    def sctpCapabilities(self) -> Optional[SctpCapabilities]:
        if self._device:
            return self._device.sctpCapabilities
        else:
            raise Exception('device has not loaded')

    def createSendTransport(self, transportId: str, iceParameters, iceCandidates, dtlsParameters, sctpParameters,
                            onConnectFunc, onProduceFunc, onProduceDataFunc):
        self._sendTransport = self._device.createSendTransport(
            id=transportId,
            iceParameters=iceParameters,
            iceCandidates=iceCandidates,
            dtlsParameters=dtlsParameters,
            sctpParameters=sctpParameters
        )

        @self._sendTransport.on('connect')
        async def onConnect(inputDtlsParameters):
            await onConnectFunc(inputDtlsParameters)

        @self._sendTransport.on('produce')
        async def onProduce(kind: str, rtpParameters, appData: dict):
            return await onProduceFunc(kind, rtpParameters, appData)

        @self._sendTransport.on('producedata')
        async def onProduceData(
                sctpStreamParameters: SctpStreamParameters,
                label: str,
                protocol: str,
                appData: dict
        ):
            return await onProduceDataFunc(sctpStreamParameters, label, protocol, appData)

    @property
    def sendTransportId(self):
        if self._sendTransport:
            return self._sendTransport.id
        else:
            return None

    def createRecvTransport(self, transportId: str, iceParameters, iceCandidates, dtlsParameters, sctpParameters,
                            onConnectFunc):
        self._recvTransport = self._device.createRecvTransport(
            id=transportId,
            iceParameters=iceParameters,
            iceCandidates=iceCandidates,
            dtlsParameters=dtlsParameters,
            sctpParameters=sctpParameters
        )

        @self._recvTransport.on('connect')
        async def onConnect(inputDtlsParameters):
            await onConnectFunc(inputDtlsParameters)

    @property
    def recvTransportId(self):
        if self._recvTransport:
            return self._recvTransport.id
        else:
            return None

    async def produce(self):
        if self._videoTrack:
            videoProducer: Producer = await self._sendTransport.produce(
                track=self._videoTrack,
                stopTracks=False,
                appData={}
            )
            self._producers.append(videoProducer)
        if self._audioTrack:
            audioProducer: Producer = await self._sendTransport.produce(
                track=self._audioTrack,
                stopTracks=False,
                appData={}
            )
            self._producers.append(audioProducer)

    def _generateRecordFilePath(self, mePeer: Peer, consumerId: str, producePeer: Peer, producerId: str,
                                kind: Literal['audio', 'video']) -> tuple:
        peerId = producePeer.peerId
        displayName = producePeer.data.displayName
        roomId = mePeer.room.roomId
        suffix: str
        if kind == 'audio':
            suffix = 'mp3'
        else:  # kind == 'video'
            suffix = 'mp4'
        return roomId, f'{displayName}({peerId})_{kind}({consumerId})', suffix

    async def consume(self, mePeer: Peer, consumerId: str,
                      producePeer: Peer, producerId: str, kind: Literal['audio', 'video'], rtpParameters: dict):
        recorder: Union[MediaBlackhole, MediaRecorder]
        if self._recordDirectoryPath == '':
            recorder = MediaBlackhole()
        else:
            relativeDirectory, fileName, suffix = self._generateRecordFilePath(mePeer, consumerId, producePeer,
                                                                               producerId,
                                                                               kind)
            recordFileParentPath = self._recordDirectoryPath + '/' + relativeDirectory
            if not os.path.exists(recordFileParentPath):
                os.makedirs(recordFileParentPath)
            recordFilePath = recordFileParentPath + '/' + f'{fileName}.{suffix}'
            recorder = MediaRecorder(file=recordFilePath)
        self._recorders[producePeer.peerId] = {consumerId: recorder}

        consumer: Consumer = await self._recvTransport.consume(
            id=consumerId,
            producerId=producerId,
            kind=kind,
            rtpParameters=rtpParameters
        )
        self._consumers.append(consumer)
        recorder.addTrack(consumer.track)
        await recorder.start()

    async def consumeData(self, dataConsumerId, dataProducerId, sctpStreamParameters, label, protocol, appData,
                          onMessageFunc):
        dataConsumer: DataConsumer = await self._recvTransport.consumeData(
            id=dataConsumerId,
            dataProducerId=dataProducerId,
            sctpStreamParameters=sctpStreamParameters,
            label=label,
            protocol=protocol,
            appData=appData
        )
        self._dataConsumers.append(dataConsumer)

        @dataConsumer.on('message')
        def onMessage(recvMessage):
            onMessageFunc(recvMessage)

    async def close(self, stopTaskLoopFunc):
        for consumer in self._consumers:
            await consumer.close()
        for dataConsumer in self._dataConsumers:
            await dataConsumer.close()
        for producer in self._producers:
            await producer.close()

        stopTaskLoopFunc()

        if self._sendTransport:
            await self._sendTransport.close()
        if self._recvTransport:
            await self._recvTransport.close()

        for consumerIdToRecorderEntry in self._recorders.values():
            for recorder in consumerIdToRecorderEntry.values():
                await recorder.stop()

        # no MediaPlayer.stop() api
        # await self._player.stop()



import sys
if sys.version_info >= (3, 8):
    from typing import List, Optional, Literal, Dict
else:
    from typing import List, Optional, Dict
    from typing_extensions import Literal

import logging
from pydantic import BaseModel
from aiortc import RTCIceCandidate
from ...deps.sdp_transform import sdp_transform
from .media_section import MediaSection, AnswerMediaSection, OfferMediaSection
from ...models.transport import IceCandidate, IceParameters, DtlsParameters, DtlsRole, PlainRtpParameters, DtlsRole
from ...producer import ProducerCodecOptions
from ...rtp_parameters import MediaKind, RtpParameters
from ...sctp_parameters import SctpParameters


class MediaSectionIdx(BaseModel):
    idx: int
    reuseMid: Optional[str]

class RemoteSdp:
    def __init__(
        self,
        iceParameters: Optional[IceParameters] = None,
        iceCandidates: List[IceCandidate] = [],
        dtlsParameters: Optional[DtlsParameters] = None,
        sctpParameters: SctpParameters = None,
        plainRtpParameters: Optional[PlainRtpParameters] = None,
        planB: bool = False
    ):
        self._iceParameters: Optional[IceParameters] = iceParameters
        self._iceCandidates: List[RTCIceCandidate] = iceCandidates
        self._dtlsParameters: Optional[DtlsParameters] = dtlsParameters
        self._sctpParameters: Optional[SctpParameters] = sctpParameters
        self._plainRtpParameters: Optional[PlainRtpParameters] = plainRtpParameters
        self._planB: bool = planB
        self._mediaSections: List[MediaSection] = []
        self._midToIndex: Dict[str, int] = {}
        self._firstMid: Optional[str] = None
        self._sdpDict: dict = {
            'version': 0,
            'origin': {
                'address': '0.0.0.0',
                'ipVer': 4,
                'netType': 'IN',
                'sessionId': 10000,
                'sessionVersion': 0,
                'username': 'mediasoup-client'
            },
            'name': '-',
            'timing': { 'start': 0, 'stop': 0 },
            'media': []
        }

        # If ICE parameters are given, add ICE-Lite indicator.
        if iceParameters:
            if iceParameters.iceLite:
                self._sdpDict['icelite'] = 'ice-lite'
        
        # If DTLS parameters are given, assume WebRTC and BUNDLE.
        if dtlsParameters:
            self._sdpDict['msidSemantic'] = { 'semantic': 'WMS', 'token': '*' }
            # NOTE: aiortc currently only support sha-256
            for fingerprint in dtlsParameters.fingerprints:
                if fingerprint.algorithm == 'sha-256':

                    self._sdpDict['fingerprint'] = {
                        'type': fingerprint.algorithm,
                        'hash': fingerprint.value
                    }
            # # NOTE: Mediasoup Client take the latest fingerprint.
            # self._sdpDict['fingerprint'] = {
            #     'type': dtlsParameters.fingerprints[-1].algorithm,
            #     'hash': dtlsParameters.fingerprints[-1].value
            # }
            self._sdpDict['groups'] = [{ 'type': 'BUNDLE', 'mids': '' }]
        
        # If there are plain RPT parameters, override SDP origin.
        if plainRtpParameters:
            self._sdpDict['origin']['address'] = plainRtpParameters.ip
            self._sdpDict['origin']['ipVer'] = plainRtpParameters.ipVersion
    
    def updateIceParameters(self, iceParameters: IceParameters):
        logging.debug(f'updateIceParameters() [iceParameters:{iceParameters}]')
        self._iceParameters = iceParameters
        self._sdpDict['icelite'] = 'ice-lite' if iceParameters.iceLite else None
    
    def updateDtlsRole(self, role: DtlsRole):
        logging.debug(f'updateDtlsRole() [role:{role}]')
        if self._dtlsParameters:
            self._dtlsParameters.role = role
            for mediaSection in self._mediaSections:
                mediaSection.setDtlsRole(role)
        
    def getNextMediaSectionIdx(self):
        # If a closed media section is found, return its index.
        for idx, mediaSection in enumerate(self._mediaSections):
            if mediaSection.closed:
                logging.debug(f'remoteSdp | getNextMediaSectionIdx() Closed media sections found { mediaSection}')
                return MediaSectionIdx(idx=idx, reuseMid=mediaSection.mid)
        # If no closed media section is found, return next one.
        logging.debug(f'remoteSdp | getNextMediaSectionIdx() No closed media sections found, return next {len(self._mediaSections)}')
        return MediaSectionIdx(idx=len(self._mediaSections))
    
    def send(
        self,
        offerMediaDict: dict,
        offerRtpParameters: RtpParameters,
        answerRtpParameters: RtpParameters,
        codecOptions: Optional[ProducerCodecOptions]=None,
        reuseMid: Optional[str]=None,
        extmapAllowMixed = False
    ):
        logging.debug(f'remoteSdp | send() offerMediaDict {offerMediaDict}')
        mediaSection = AnswerMediaSection(
            sctpParameters=self._sctpParameters,
            iceParameters=self._iceParameters,
            iceCandidates=self._iceCandidates,
            dtlsParameters=self._dtlsParameters,
            plainRtpParameters=self._plainRtpParameters,
            planB=self._planB,
            offerMediaDict=offerMediaDict,
            offerRtpParameters=offerRtpParameters,
            answerRtpParameters=answerRtpParameters,
            codecOptions=codecOptions,
            extmapAllowMixed=extmapAllowMixed
        )
        # Unified-Plan with closed media section replacement.
        if reuseMid:
            self._replaceMediaSection(mediaSection, reuseMid)
        # Unified-Plan or Plan-B with different media kind.
        elif mediaSection.mid not in self._midToIndex.keys():
            self._addMediaSection(mediaSection)
        # Plan-B with same media kind.
        else:
            self._replaceMediaSection(mediaSection)
        
    def receive(
        self,
        mid: str,
        kind: Literal['audio', 'video', 'application'],
        offerRtpParameters: RtpParameters,
        streamId: str,
        trackId: str
    ):
        idx: int = self._midToIndex.get(str(mid), -1)
        if idx != -1:
            mediaSection = self._mediaSections[idx]
            # Plan-B.
            mediaSection.planBReceive(offerRtpParameters, streamId, trackId)
            self._replaceMediaSection(mediaSection)
        # Unified-Plan or different media kind.
        else:
            mediaSection = OfferMediaSection(
                iceParameters=self._iceParameters,
                iceCandidates=self._iceCandidates,
                dtlsParameters=self._dtlsParameters,
                plainRtpParameters=self._plainRtpParameters,
                planB=self._planB,
                mid=mid,
                kind=kind,
                offerRtpParameters=offerRtpParameters,
                streamId=streamId,
                trackId=trackId
            )
            # Let's try to recycle a closed media section (if any).
            # NOTE: Yes, we can recycle a closed m=audio section with a new m=video.
            closedMediaSections = [ms for ms in self._mediaSections if ms.closed]
            if closedMediaSections:
                self._replaceMediaSection(mediaSection, closedMediaSections[0].mid)
            else:
                self._addMediaSection(mediaSection)
            
    def disableMediaSection(self, mid: str):
        idx: int = self._midToIndex.get(str(mid), -1)
        if idx == -1:
            raise Exception(f"no media section found with mid '{mid}'")
        mediaSection = self._mediaSections[idx]
        mediaSection.disable()
    
    def closeMediaSection(self, mid: str):
        idx: int = self._midToIndex.get(str(mid), -1)
        if idx == -1:
            raise Exception(f"no media section found with mid '{mid}'")
        mediaSection = self._mediaSections[idx]
        # NOTE: Closing the first m section is a pain since it invalidates the
        # bundled transport, so let's avoid it.
        if mid == self._firstMid:
            logging.debug(f'closeMediaSection() | cannot close first media section, disabling it instead [mid:{mid}]')
            self.disableMediaSection(mid)
            return
        mediaSection.close()
        # Regenerate BUNDLE mids.
        self._regenerateBundleMids()
    
    def planBStopReceiving(self, mid: str, offerRtpParameters: RtpParameters):
        idx: int = self._midToIndex.get(str(mid), -1)
        if idx == -1:
            raise Exception(f"no media section found with mid '{mid}'")
        mediaSection = self._mediaSections[idx]
        mediaSection.planBStopReceiving(offerRtpParameters)
        self._replaceMediaSection(mediaSection)
    
    def sendSctpAssociation(self, offerMediaDict: dict):
        mediaSection = AnswerMediaSection(
            iceParameters=self._iceParameters,
            iceCandidates=self._iceCandidates,
            dtlsParameters=self._dtlsParameters,
            sctpParameters=self._sctpParameters,
            plainRtpParameters=self._plainRtpParameters,
            offerMediaDict=offerMediaDict
        )
        self._addMediaSection(mediaSection)
    
    def receiveSctpAssociation(self, oldDataChannelSpec: bool = False):
        mediaSection = OfferMediaSection(
            iceParameters=self._iceParameters,
            iceCandidates=self._iceCandidates,
            dtlsParameters=self._dtlsParameters,
            sctpParameters=self._sctpParameters,
            plainRtpParameters=self._plainRtpParameters,
            mid='datachannel',
            kind='application',
            oldDataChannelSpec=oldDataChannelSpec
        )
        self._addMediaSection(mediaSection)
    
    def getSdp(self) -> str:
        # Increase SDP version.
        self._sdpDict['origin']['sessionVersion'] += 1
        return sdp_transform.write(self._sdpDict)
    
    def _addMediaSection(self, newMediaSection: MediaSection):
        if self._firstMid == None:
            self._firstMid = newMediaSection.mid
        # Add to the list.
        self._mediaSections.append(newMediaSection)
        # Add to the map.
        self._midToIndex[newMediaSection.mid] = len(self._mediaSections) - 1
        # Add to the SDP object.
        self._sdpDict['media'].append(newMediaSection.getDict())
        # Regenerate BUNDLE mids.
        self._regenerateBundleMids()

    def _replaceMediaSection(self, newMediaSection: MediaSection, reuseMid: Optional[str]=None):
        # Store it in the map.
        if reuseMid:
            idx = self._midToIndex.get(str(reuseMid), -1)
            if idx == -1:
                raise Exception(f"no media section found with mid '{reuseMid}'")
            oldMediaSection = self._mediaSections[idx]
            # Replace the index in the vector with the new media section.
            self._mediaSections[idx] = newMediaSection
            # Update the map.
            del self._midToIndex[oldMediaSection.mid]
            self._midToIndex[newMediaSection.mid] =idx
            # Update the SDP object.
            self._sdpDict['media'][idx] = newMediaSection.getDict()
            # Regenerate BUNDLE mids.
            self._regenerateBundleMids()
        else:
            idx = self._midToIndex.get(newMediaSection.mid, -1)
            if idx == -1:
                raise Exception(f"no media section found with mid '{newMediaSection.mid}'")
            # Replace the index in the vector with the new media section.
            self._mediaSections[idx] = newMediaSection
            # Update the SDP object.
            self._sdpDict['media'][idx] = newMediaSection.getDict()
    
    def _regenerateBundleMids(self):
        if not self._dtlsParameters:
            return
        self._sdpDict['groups'][0]['mids'] = ' '.join([ms.mid for ms in self._mediaSections if not ms.closed])



import re
from pydantic import BaseModel


class ScalabilityMode(BaseModel):
    spatialLayers: int
    temporalLayers: int

reg = r'^[LS]([1-9]\\d{0,1})T([1-9]\\d{0,1})'
def parse(scalabilityMode: str) -> ScalabilityMode:
    match = re.match(reg, scalabilityMode)
    if match:
        return ScalabilityMode(
            spatialLayers=int(match[1]),
            temporalLayers=int(match[2])
        )
    else:
        return ScalabilityMode(
            spatialLayers=1,
            temporalLayers=1
        )


# todo: my design hasn't finished yet，(⊙﹏⊙)
# local：collect statistics from local webrtc device, e.g. browser
# remote：collect statistics from mediasoup server side
class StatsCollector:
    def __init__(self):
        pass

    def onGetSendTransportRemoteStats(self):
        pass

    def onGetSendTransportLocalStats(self):
        pass

    def onGetRecvTransportRemoteStats(self):
        pass

    def onGetRecvTransportLocalStats(self):
        pass

    def onGetAudioRemoteStats(self):
        # onGetProducerStats
        pass

    def onGetAudioLocalStats(self):
        pass

    def onGetVideoRemoteStats(self):
        # onGetProducerStats
        pass

    def onGetVideoLocalStats(self):
        pass

    def onGetConsumerLocalStats(self):
        pass

    def onGetConsumerRemoteStats(self):
        pass

    def onGetDataProducerRemoteStats(self):
        pass

    def onGetDataConsumerRemoteStats(self):
        pass



import sys
if sys.version_info >= (3, 8):
    from typing import Dict, Literal, List, Optional, Any
else:
    from typing import Dict, List, Optional, Any
    from typing_extensions import Literal

import logging
from aiortc import RTCIceServer, RTCPeerConnection, RTCSessionDescription, RTCRtpTransceiver, MediaStreamTrack
from ..deps.sdp_transform import sdp_transform
from .sdp import common_utils
from .sdp.remote_sdp import RemoteSdp
from .sdp.unified_plan_utils import addLegacySimulcast, getRtpEncodings
from .sdp.common_utils import applyCodecParameters, extractDtlsParameters
from .handler_interface import HandlerInterface
from ..ortc import ExtendedRtpCapabilities
from ..rtp_parameters import MediaKind, RtpParameters, RtpCapabilities, RtpCodecCapability, RtpEncodingParameters, RtcpParameters
from ..sctp_parameters import SctpCapabilities, SctpParameters, SctpStreamParameters
from ..ortc import getSendingRtpParameters, getSendingRemoteRtpParameters, reduceCodecs
from ..scalability_modes import parse as smParse
from ..models.transport import IceCandidate, IceParameters, DtlsParameters, DtlsRole
from ..models.handler_interface import HandlerRunOptions, HandlerSendOptions, HandlerSendResult, HandlerSendDataChannelResult, HandlerReceiveDataChannelResult, HandlerReceiveOptions, HandlerReceiveResult, HandlerReceiveDataChannelOptions
from ..producer import ProducerCodecOptions


SCTP_NUM_STREAMS = { 'OS': 1024, 'MIS': 1024 }

class AiortcHandler(HandlerInterface):

    def __init__(self, tracks: List[MediaStreamTrack]=[], loop=None):
        super(AiortcHandler, self).__init__(loop=loop)
        # Handler direction.
        self._direction: Optional[Literal['send', 'recv']] = None
        # Remote SDP handler.
        self._remoteSdp: Optional[RemoteSdp] = None
        # Generic sending RTP parameters for audio and video.
        self._sendingRtpParametersByKind: Dict[str, RtpParameters] = {}
        # Generic sending RTP parameters for audio and video suitable for the SDP
        # remote answer.
        self._sendingRemoteRtpParametersByKind: Dict[str, RtpParameters] = {}
        # RTCPeerConnection instance.
        self._pc: Optional[RTCPeerConnection] = None
        # Map of RTCTransceivers indexed by MID.
        self._mapMidTransceiver: Dict[str, RTCRtpTransceiver] = {}
        # Whether a DataChannel m=application section has been created.
        self._hasDataChannelMediaSection = False
        # Sending DataChannel id value counter. Incremented for each new DataChannel.
        self._nextSendSctpStreamId = 0
        # Got transport local and remote parameters.
        self._transportReady = False
        self._tracks = tracks

    @classmethod
    def createFactory(cls, tracks: List[MediaStreamTrack]=[], loop=None):
        return lambda: cls(tracks, loop)

    @property
    def name(self) -> str:
        return 'aiortc'
    
    @property
    def pc(self) -> RTCPeerConnection:
        if self._pc:
            return self._pc
        else:
            raise Exception('PeerConnection not ready')
    
    @property
    def remoteSdp(self) -> RemoteSdp:
        if self._remoteSdp:
            return self._remoteSdp
        else:
            raise Exception('Remote SDP not ready')
    
    async def close(self):
        logging.debug('close()')

        if self._pc:
            await self._pc.close()

    async def getNativeRtpCapabilities(self) -> RtpCapabilities:
        logging.debug('getNativeRtpCapabilities()')

        pc = RTCPeerConnection()
        for track in self._tracks:
            pc.addTrack(track)
        pc.addTransceiver('audio')
        pc.addTransceiver('video')

        offer: RTCSessionDescription = await pc.createOffer()
        await pc.close()

        sdpDict: dict = sdp_transform.parse(offer.sdp)
        nativeRtpCapabilities:RtpCapabilities = common_utils.extractRtpCapabilities(sdpDict)

        return nativeRtpCapabilities
    
    async def getNativeSctpCapabilities(self) -> SctpCapabilities:
        logging.debug('getNativeSctpCapabilities()')
        return SctpCapabilities.parse_obj({
            'numStreams': SCTP_NUM_STREAMS
        })
    
    def run(
        self,
        direction: Literal['send', 'recv'],
        iceParameters: IceParameters,
        iceCandidates: List[IceCandidate],
        dtlsParameters: DtlsParameters,
        extendedRtpCapabilities: ExtendedRtpCapabilities,
        sctpParameters: Optional[SctpParameters]=None,
        iceServers: Optional[RTCIceServer]=None,
        iceTransportPolicy: Optional[Literal['all', 'relay']]=None,
        additionalSettings: Optional[Any]=None,
        proprietaryConstraints: Optional[Any]=None
    ):
        logging.debug('AiortcHandler run()')
        options = HandlerRunOptions(
            direction=direction,
            iceParameters=iceParameters,
            iceCandidates=iceCandidates,
            dtlsParameters=dtlsParameters,
            sctpParameters=sctpParameters,
            iceServers=iceServers,
            iceTransportPolicy=iceTransportPolicy,
            additionalSettings=additionalSettings,
            proprietaryConstraints=proprietaryConstraints,
            extendedRtpCapabilities=extendedRtpCapabilities
        )
        self._direction = options.direction
        self._remoteSdp = RemoteSdp(
            iceParameters=options.iceParameters,
            iceCandidates=options.iceCandidates,
            dtlsParameters=options.dtlsParameters,
            sctpParameters=options.sctpParameters
        )
        self._sendingRtpParametersByKind = {
            'audio': getSendingRtpParameters('audio', options.extendedRtpCapabilities),
            'video': getSendingRtpParameters('video', options.extendedRtpCapabilities)
        }
        self._sendingRemoteRtpParametersByKind = {
            'audio': getSendingRemoteRtpParameters('audio', options.extendedRtpCapabilities),
            'video': getSendingRemoteRtpParameters('video', options.extendedRtpCapabilities)
        }
        self._pc = RTCPeerConnection()

        @self._pc.on('iceconnectionstatechange')
        def on_iceconnectionstatechange():
            if self._pc.iceConnectionState == 'checking':
                self.emit('@connectionstatechange', 'connecting')
            elif self._pc.iceConnectionState in ['connected', 'completed']:
                self.emit('@connectionstatechange', 'connected')
            elif self._pc.iceConnectionState == 'failed':
                self.emit('@connectionstatechange', 'failed')
            elif self._pc.iceConnectionState == 'disconnected':
                self.emit('@connectionstatechange', 'disconnected')
            elif self._pc.iceConnectionState == 'closed':
                self.emit('@connectionstatechange', 'closed')
        
    async def updateIceServers(self, iceServers):
        logging.warning('updateIceServers() not implemented')
        # TODO: aiortc can not update iceServers
    
    async def restartIce(self, iceParameters):
        logging.debug('restartIce()')
        self._remoteSdp.updateIceParameters(iceParameters)
        if not self._transportReady:
            return
        if self._direction == 'send':
            # NOTE: aiortc RTCPeerConnection createOffer do not have iceRestart options
            offer = await self._pc.createOffer()
            logging.debug(f'restartIce() | calling pc.setLocalDescription() [offer:{offer}]')
            await self._pc.setLocalDescription(offer)
            answer: RTCSessionDescription = RTCSessionDescription(
                type='answer',
                sdp=self._remoteSdp.getSdp()
            )
            logging.debug(f'restartIce() | calling pc.setRemoteDescription() [answer:{answer}]')
            await self._pc.setRemoteDescription(answer)
        else:
            offer: RTCSessionDescription = RTCSessionDescription(
                type='offer',
                sdp=self._remoteSdp.getSdp()
            )
            logging.debug(f'restartIce() | calling pc.setRemoteDescription() [offer:{offer}]')
            await self._pc.setRemoteDescription(offer)
            answer = await self._pc.createAnswer()
            logging.debug(f'restartIce() | calling pc.setLocalDescription() [answer:{answer}]')
            await self._pc.setLocalDescription(answer)
        
    async def getTransportStats(self):
        return self._pc.getStats()

    async def send(
        self,
        track: MediaStreamTrack,
        encodings: List[RtpEncodingParameters]=[],
        codecOptions: Optional[ProducerCodecOptions]=None,
        codec: Optional[RtpCodecCapability]=None
    ) -> HandlerSendResult:
        options = HandlerSendOptions(
            track=track,
            encodings=encodings,
            codecOptions=codecOptions,
            codec=codec
        )
        self._assertSendDirection()
        logging.debug(f'send() [kind:{options.track.kind}, track.id:{options.track.id}]')
        if options.encodings:
            for idx in range(len(options.encodings)):
                options.encodings[idx].rid = f'r{idx}'

        sendingRtpParameters: RtpParameters = self._sendingRtpParametersByKind[options.track.kind].copy(deep=True)
        sendingRtpParameters.codecs = reduceCodecs(sendingRtpParameters.codecs, options.codec)

        sendingRemoteRtpParameters: RtpParameters = self._sendingRemoteRtpParametersByKind[options.track.kind].copy(deep=True)
        sendingRemoteRtpParameters.codecs = reduceCodecs(sendingRemoteRtpParameters.codecs, options.codec)

        mediaSectionIdx = self.remoteSdp.getNextMediaSectionIdx()
        transceiver = self.pc.addTransceiver(options.track, direction='sendonly')

        offer: RTCSessionDescription  = await self.pc.createOffer()
        offerMediaDict: dict
        localSdpDict = sdp_transform.parse(offer.sdp)
        if not self._transportReady:
            await self._setupTransport(localDtlsRole='server', localSdpDict=localSdpDict)
        # Special case for VP9 with SVC.
        hackVp9Svc = False
        if options.encodings:
            layers=smParse(options.encodings[0].scalabilityMode if options.encodings[0].scalabilityMode else '')
        else:
            layers=smParse('')
        if len(options.encodings) == 1 and layers.spatialLayers > 1 and sendingRtpParameters.codecs[0].mimeType.lower() == 'video/vp9':
            logging.debug('send() | enabling legacy simulcast for VP9 SVC')
            hackVp9Svc = True
            localSdpDict = sdp_transform.parse
            offerMediaDict = localSdpDict['media'][mediaSectionIdx.idx]
            addLegacySimulcast(offerMediaDict=offerMediaDict, numStreams=layers.spatialLayers)
            offer = RTCSessionDescription(
                type='offer',
                sdp=sdp_transform.write(localSdpDict)
            )
        
        logging.debug(f'send() | calling pc.setLocalDescription() [offer:{offer}]')

        await self.pc.setLocalDescription(offer)
        # We can now get the transceiver.mid.
        localId = transceiver.mid
        # Set MID.
        sendingRtpParameters.mid = localId
        localSdpDict = sdp_transform.parse(self.pc.localDescription.sdp)

        offerMediaDict = localSdpDict['media'][mediaSectionIdx.idx]

        logging.debug(f"send() | get offerMediaDict {offerMediaDict} \n from localSdpDict {localSdpDict['media']} index {mediaSectionIdx.idx}")
        # Set RTCP CNAME.
        if sendingRtpParameters.rtcp == None:
            sendingRtpParameters.rtcp = RtcpParameters()
        sendingRtpParameters.rtcp.cname = common_utils.getCname(offerMediaDict)
        # Set RTP encodings by parsing the SDP offer if no encodings are given.
        if not options.encodings:
            sendingRtpParameters.encodings = getRtpEncodings(offerMediaDict)
        # Set RTP encodings by parsing the SDP offer and complete them with given
        # one if just a single encoding has been given.
        elif len(options.encodings) == 1:
            newEncodings = getRtpEncodings(offerMediaDict)
            if newEncodings and options.encodings[0]:
                firstEncodingDict: dict = newEncodings[0].dict()
                optionsEncodingDict: dict = options.encodings[0].dict()
                firstEncodingDict.update(optionsEncodingDict)
                newEncodings[0] = RtpEncodingParameters(**firstEncodingDict)
                if hackVp9Svc:
                    newEncodings = [newEncodings[0]]
            sendingRtpParameters.encodings = newEncodings
        # Otherwise if more than 1 encoding are given use them verbatim.
        else:
            sendingRtpParameters.encodings = options.encodings
        # If VP8 or H264 and there is effective simulcast, add scalabilityMode to
        # each encoding.
        if len(sendingRtpParameters.encodings) > 1 and (sendingRtpParameters.codecs[0].mimeType.lower() == 'video/vp8' or sendingRtpParameters.codecs[0].mimeType.lower() == 'video/h264'):
            for encoding in sendingRtpParameters.encodings:
                encoding.scalabilityMode = 'S1T3'
        self.remoteSdp.send(
            offerMediaDict=offerMediaDict,
            reuseMid=mediaSectionIdx.reuseMid,
            offerRtpParameters=sendingRtpParameters,
            answerRtpParameters=sendingRemoteRtpParameters,
            codecOptions=options.codecOptions,
            extmapAllowMixed=True
        )
        answer: RTCSessionDescription = RTCSessionDescription(
            type='answer',
            sdp=self.remoteSdp.getSdp()
        )
        logging.debug(f'send() | calling pc.setRemoteDescription() [answer:{answer}]')
        await self.pc.setRemoteDescription(answer)
        # Store in the map.
        self._mapMidTransceiver[localId] = transceiver
        return HandlerSendResult(
            localId=localId,
            rtpParameters=sendingRtpParameters,
            rtpSender=transceiver.sender
        )

    async def stopSending(self, localId):
        pass
        # self._assertSendDirection()
        # logging.debug(f'stopSending() [localId:{localId}]')
        # transceiver = self._mapMidTransceiver.get(localId)
        # if not transceiver:
        #     raise Exception('associated RTCRtpTransceiver not found')
        # transceiver.sender.replaceTrack()
        # # TODO:RTCPeerConnection do not have removeTrack()
        # self._pc.removeTrack(transceiver.sender)
        # self._remoteSdp.closeMediaSection(transceiver.mid)
        # offer = await self._pc.createOffer()
        # logging.debug(f'stopSending() | calling pc.setLocalDescription() [offer:{offer}]')
        # await self._pc.localDescription(offer)
        # answer: RTCSessionDescription = RTCSessionDescription(
        #     type='answer',
        #     sdp=self._remoteSdp.getSdp()
        # )
        # logging.debug(f'stopSending() | calling pc.setRemoteDescription() [answer:{answer}]')
        # await self._pc.setRemoteDescription(answer)
    
    async def replaceTrack(self, localId, track=None):
        self._assertSendDirection()
        if track:
            logging.debug(f'replaceTrack() [localId:{localId}, track.id:{track.id}]')
        else:
            logging.debug(f'replaceTrack() [localId:{localId}, no track]')
        transceiver = self._mapMidTransceiver.get(localId)
        if not transceiver:
            raise Exception('associated RTCRtpTransceiver not found')

        await transceiver.sender.replaceTrack(track)
    
    async def setMaxSpatialLayer(self, localId: str, spatialLayer: int):
        logging.warning('setMaxSpatialLayer() not implemented')
        # NOTE: RTCRtpSender do not have getParameters()
        # self._assertSendDirection()
        # logging.debug(f'setMaxSpatialLayer() [localId:{localId}, spatialLayer:{spatialLayer}]')
        # transceiver = self._mapMidTransceiver.get(localId)
        # if not transceiver:
        #     raise Exception('associated RTCRtpTransceiver not found')
        # parameters = transceiver.sender.getParameters()
    
    async def setRtpEncodingParameters(self, localId: str, params: Any):
        logging.warning('setRtpEncodingParameters() not implemented')
        # NOTE: RTCRtpSender do not have getParameters()
    
    async def getSenderStats(self, localId: str):
        self._assertSendDirection()
        transceiver = self._mapMidTransceiver.get(localId)
        if not transceiver:
            raise Exception('associated RTCRtpTransceiver not found')
        return await transceiver.sender.getStats()
    
    async def sendDataChannel(
        self,
        streamId: Optional[int]=None,
        ordered: Optional[bool]=True,
        maxPacketLifeTime: Optional[int]=None,
        maxRetransmits: Optional[int]=None,
        priority: Optional[Literal['very-low','low','medium','high']]=None,
        label: Optional[str]=None,
        protocol: Optional[str]=None
    ) -> HandlerSendDataChannelResult:
        if streamId == None:
            streamId = self._nextSendSctpStreamId
        options=SctpStreamParameters(
            streamId=streamId,
            ordered=ordered,
            maxPacketLifeTime=maxPacketLifeTime,
            maxRetransmits=maxRetransmits,
            priority=priority,
            label=label,
            protocol=protocol
        )
        self._assertSendDirection()
        logging.debug('sendDataChannel()')
        dataChannel = self.pc.createDataChannel(
            label=options.label,
            maxPacketLifeTime=options.maxPacketLifeTime,
            ordered=options.ordered,
            protocol=options.protocol,
            negotiated=True,
            id=self._nextSendSctpStreamId
        )
        # Increase next id.
        self._nextSendSctpStreamId = (self._nextSendSctpStreamId + 1) % SCTP_NUM_STREAMS.get('MIS', 1)
        # If this is the first DataChannel we need to create the SDP answer with
        # m=application section.
        if not self._hasDataChannelMediaSection:
            offer: RTCSessionDescription = await self.pc.createOffer()
            localSdpDict = sdp_transform.parse(offer.sdp)
            offerMediaDicts = [m for m in localSdpDict.get('media') if m.get('type') == 'application']
            if not offerMediaDicts:
                raise Exception('No datachannel')
            offerMediaDict = offerMediaDicts[0]

            if not self._transportReady:
                await self._setupTransport(localDtlsRole='server', localSdpDict=localSdpDict)
            
            logging.debug(f'sendDataChannel() | calling pc.setLocalDescription() [offer:{offer}]')
            await self.pc.setLocalDescription(offer)
            self.remoteSdp.sendSctpAssociation(offerMediaDict=offerMediaDict)
            answer: RTCSessionDescription = RTCSessionDescription(
                type='answer',
                sdp=self.remoteSdp.getSdp()
            )

            logging.debug(f'sendDataChannel() | calling pc.setRemoteDescription() [answer:{answer}]')
            await self.pc.setRemoteDescription(answer)
            self._hasDataChannelMediaSection = True
        
        return HandlerSendDataChannelResult(
            dataChannel=dataChannel,
            sctpStreamParameters=options
        )
    
    async def receive(
        self,
        trackId: str,
        kind: MediaKind,
        rtpParameters: RtpParameters
    ) -> HandlerReceiveResult:
        options = HandlerReceiveOptions(
            trackId=trackId,
            kind=kind,
            rtpParameters=rtpParameters
        )
        self._assertRecvDirection()
        logging.debug(f'receive() [trackId:{options.trackId}, kind:{options.kind}]')
        localId = options.rtpParameters.mid if options.rtpParameters.mid != None else str(len(self._mapMidTransceiver))
        self.remoteSdp.receive(
            mid=localId,
            kind=options.kind,
            offerRtpParameters=options.rtpParameters,
            streamId=options.rtpParameters.rtcp.cname,
            trackId=options.trackId
        )
        offer: RTCSessionDescription = RTCSessionDescription(
            type='offer',
            sdp=self.remoteSdp.getSdp()
        )
        logging.debug(f'receive() | calling pc.setRemoteDescription() [offer:{offer}]')
        await self.pc.setRemoteDescription(offer)
        answer: RTCSessionDescription = await self.pc.createAnswer()
        localSdpDict = sdp_transform.parse(answer.sdp)
        answerMediaDict = [m for m in localSdpDict.get('media') if str(m.get('mid')) == localId][0]
        # May need to modify codec parameters in the answer based on codec
        # parameters in the offer.
        applyCodecParameters(offerRtpParameters=options.rtpParameters, answerMediaDict=answerMediaDict)
        answer = RTCSessionDescription(
            type='answer',
            sdp=sdp_transform.write(localSdpDict)
        )
        if not self._transportReady:
            await self._setupTransport(localDtlsRole='client', localSdpDict=localSdpDict)
        logging.debug(f'receive() | calling pc.setLocalDescription() [answer:{answer}]')
        await self.pc.setLocalDescription(answer)
        transceivers = [t for t in self.pc.getTransceivers() if t.mid == localId]
        if not transceivers:
            raise Exception('new RTCRtpTransceiver not found')
        # Store in the map.
        transceiver = transceivers[0]
        self._mapMidTransceiver[localId] = transceiver

        return HandlerReceiveResult(
            localId=localId,
            track=transceiver.receiver.track,
            rtpReceiver=transceiver.receiver
        )
        
    async def stopReceiving(self, localId: str):
        self._assertRecvDirection()
        logging.debug(f'stopReceiving() [localId:{localId}]')
        transceiver = self._mapMidTransceiver.get(localId)
        if not transceiver:
            raise Exception('associated RTCRtpTransceiver not found')
        self.remoteSdp.closeMediaSection(transceiver.mid)
        offer: RTCSessionDescription = RTCSessionDescription(
            type='offer',
            sdp=self.remoteSdp.getSdp()
        )
        logging.debug(f'stopReceiving() | calling pc.setRemoteDescription() [offer:{offer}]')
        await self.pc.setRemoteDescription(offer)
        answer = await self.pc.createAnswer()
        logging.debug(f'stopReceiving() | calling pc.setLocalDescription() [answer:{answer}]')
        await self.pc.setLocalDescription(answer)
    
    async def getReceiverStats(self, localId: str):
        self._assertRecvDirection()
        transceiver = self._mapMidTransceiver.get(localId)
        if not transceiver:
            raise Exception('associated RTCRtpTransceiver not found')
        return await transceiver.receiver.getStats()
    
    async def receiveDataChannel(
        self,
        sctpStreamParameters: SctpStreamParameters,
        label: Optional[str]=None,
        protocol: Optional[str]=None
    ) -> HandlerReceiveDataChannelResult:
        options = HandlerReceiveDataChannelOptions(
            sctpStreamParameters=sctpStreamParameters,
            label=label,
            protocol=protocol
        )
        self._assertRecvDirection()
        logging.debug(f'[receiveDataChannel() [options:{options.sctpStreamParameters}]]')
        dataChannel = self.pc.createDataChannel(
            label=options.label,
            maxPacketLifeTime=options.sctpStreamParameters.maxPacketLifeTime,
            maxRetransmits=options.sctpStreamParameters.maxRetransmits,
            ordered=options.sctpStreamParameters.ordered,
            protocol=options.protocol,
            negotiated=True,
            id=options.sctpStreamParameters.streamId
        )

        # If this is the first DataChannel we need to create the SDP offer with
        # m=application section.
        if not self._hasDataChannelMediaSection:
            self.remoteSdp.receiveSctpAssociation()
            offer: RTCSessionDescription = RTCSessionDescription(
                type='offer',
                sdp=self.remoteSdp.getSdp()
            )
            logging.debug(f'receiveDataChannel() | calling pc.setRemoteDescription() [offer:{offer}]')
            await self.pc.setRemoteDescription(offer)
            answer = await self.pc.createAnswer()
            if not self._transportReady:
                localSdpDict = sdp_transform.parse(answer.sdp)
                await self._setupTransport(localDtlsRole='client', localSdpDict=localSdpDict)
            logging.debug(f'receiveDataChannel() | calling pc.setRemoteDescription() [answer:{answer}]')
            await self.pc.setLocalDescription(answer)
            self._hasDataChannelMediaSection = True
        return HandlerReceiveDataChannelResult(dataChannel=dataChannel)
    
    async def _setupTransport(self, localDtlsRole: DtlsRole, localSdpDict: dict={}):
        if localSdpDict == {}:
            localSdpDict = sdp_transform.parse(self.pc.localDescription.sdp)
        # Get our local DTLS parameters.
        dtlsParameters: DtlsParameters = extractDtlsParameters(localSdpDict)
        # Set our DTLS role.
        dtlsParameters.role = localDtlsRole
        # Update the remote DTLS role in the SDP.
        self.remoteSdp.updateDtlsRole('server' if localDtlsRole == 'client' else 'client')
        # Need to tell the remote transport about our parameters.
        await self.emit_for_results('@connect', dtlsParameters)
        self._transportReady = True
    
    def _assertSendDirection(self):
        if self._direction != 'send':
            raise Exception('method can just be called for handlers with "send" direction')
    
    def _assertRecvDirection(self):
        if self._direction != 'recv':
            raise Exception('method can just be called for handlers with "recv" direction')



import logging


class Logger:
    @staticmethod
    def getLogger(enable_console, log_file_path, level):
        logger = logging.getLogger()
        logger.setLevel(level)
        # 日志格式
        formatter = logging.Formatter(
            "%(asctime)s -$- [%(threadName)s] -$- %(levelname)s -$- %(filename)s -$- %(funcName)s(%(lineno)d) -$- %(message)s")

        if enable_console:
            # 建立一个stream handler来把日志打在CMD窗口上，级别为error以上
            console_handler = logging.StreamHandler()
            console_handler.setLevel(level)
            console_handler.setFormatter(formatter)
            # 将相应的handler添加在logger对象中
            logger.addHandler(console_handler)

        if log_file_path is not None:
            # 建立一个filehandler来把日志记录在文件里，级别为debug以上
            file_handler = logging.FileHandler(log_file_path)
            file_handler.setLevel(level)
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)
        return logger



from .room_peer import Peer


class MediasoupListener:
    """
    top abstract class for ConsumerRequestListener and QueuedNotificationListener,
    with the information of current peer
    """

    def __init__(self, mePeer: Peer):
        self._mePeer = mePeer

    @property
    def mePeer(self) -> Peer:
        return self._mePeer

    @mePeer.setter
    def mePeer(self, mePeer):
        self._mePeer = mePeer



import sys
if sys.version_info >= (3, 8):
    from typing import Optional, Any, Union, Literal
else:
    from typing import Optional, Any, Union
    from typing_extensions import Literal

import logging
from pyee import AsyncIOEventEmitter
from aiortc import RTCDataChannel
from pydantic import BaseModel
from .errors import InvalidStateError
from .emitter import EnhancedEventEmitter
from .sctp_parameters import SctpStreamParameters


class DataProducerOptions(BaseModel):
    ordered: Optional[bool]
    maxPacketLifeTime: Optional[int]
    maxRetransmits: Optional[int]
    priority: Optional[Literal['very-low','low','medium','high']]
    label: Optional[str]
    protocol: Optional[str]
    appData: Optional[dict] = {}

class DataProducer(EnhancedEventEmitter):
    def __init__(
        self,
        id: str,
        dataChannel: RTCDataChannel,
        sctpStreamParameters: SctpStreamParameters,
        appData: Optional[dict] = None,
        loop=None
    ):
        super(DataProducer, self).__init__(loop=loop)

        # Closed flag.
        self._closed: bool = False
        # Observer instance.
        self._observer: AsyncIOEventEmitter = AsyncIOEventEmitter()

        self._id = id
        self._dataChannel = dataChannel
        self._sctpStreamParameters = sctpStreamParameters
        self._appData = appData
        
        self._handleDataChannel()
    
    # DataProducer id.
    @property
    def id(self) -> str:
        return self._id

    # Whether the DataProducer is closed.
    @property
    def closed(self) -> bool:
        return self._closed
    
    # SCTP stream parameters.
    @property
    def sctpStreamParameters(self) -> SctpStreamParameters:
        return self._sctpStreamParameters
    
    # DataChannel readyState.
    @property
    def readyState(self) -> Literal["closed", "closing", "connecting", "open"]:
        return self._dataChannel.readyState
    
    # DataChannel label.
    @property
    def label(self) -> str:
        return self._dataChannel.label
    
    # DataChannel protocol.
    @property
    def protocol(self) -> str:
        return self._dataChannel.protocol
    
    # DataChannel bufferedAmount.
    @property
    def bufferedAmount(self) -> int:
        return self._dataChannel.bufferedAmount
    
    # DataChannel bufferedAmountLowThreshold.
    @property
    def bufferedAmountLowThreshold(self) -> int:
        return self._dataChannel.bufferedAmountLowThreshold
    
    # Set DataChannel bufferedAmountLowThreshold.
    @bufferedAmountLowThreshold.setter
    def bufferedAmountLowThreshold(self, bufferedAmountLowThreshold: int):
        self._dataChannel.bufferedAmountLowThreshold = bufferedAmountLowThreshold
    
    # App custom data.
    @property
    def appData(self) -> Any:
        return self._appData
    
    # Invalid setter.
    @appData.setter
    def appData(self, value):
        raise Exception('cannot override appData object')
    
    # Observer.
    @property
    def observer(self) -> AsyncIOEventEmitter:
        return self._observer
    
    # Closes the DataProducer.
    async def close(self):
        if self._closed:
            return
        
        logging.debug('DataProducer close()')

        self._closed = True

        self._dataChannel.close()

        await self.emit_for_results('@close')

        # Emit observer event.
        self._observer.emit('close')
    
    # Transport was closed.
    def transportClosed(self):
        if self._closed:
            return

        logging.debug('DataProducer transportClosed()')

        self._closed = True

        self._dataChannel.close()

        self.emit('transportclose')

        self._observer.emit('close')
    
    # Send a message.
    def send(self, data: Union[bytes, str]):
        logging.debug('DataProducer send()')

        if self._closed:
            raise InvalidStateError('closed')
        
        self._dataChannel.send(data)
    
    def _handleDataChannel(self):
        @self._dataChannel.on('open')
        def on_open():
            if self._closed:
                return
            
            logging.debug('DataProducer DataChannel "open" event')

            self.emit('open')

        # NOTE: aiortc.RTCDataChannel won't emit error event, here use pyee error event
        @self._dataChannel.on('error')
        def on_error(message):
            if self._closed:
                return

            logging.error(f'DataProducer DataChannel "error" event: {message}')
            
            self.emit('error', message)

        @self._dataChannel.on('close')
        def on_close():
            if self._closed:
                return
            logging.warning('DataProducer DataChannel "close" event')
            self._closed = True
            self.emit('@close')
            self._observer.emit('close')

        @self._dataChannel.on('message')
        def on_message(message):
            if self._closed:
                return
            logging.warning(f'DataProducer DataChannel "message" event in a DataProducer, message discarded: {message}')

        @self._dataChannel.on('bufferedamountlow')
        def on_bufferedamountlow():
            if self._closed:
                return
            self.emit('bufferedamountlow')


import sys
if sys.version_info >= (3, 8):
    from typing import List, Optional, Literal
else:
    from typing import List, Optional
    from typing_extensions import Literal

import re
import logging
from aiortc import RTCIceParameters, RTCIceCandidate, RTCDtlsParameters
from ...producer import ProducerCodecOptions
from ...rtp_parameters import RtpParameters, RtpCodecParameters, RtpEncodingParameters
from ...sctp_parameters import SctpParameters
from ...models.transport import PlainRtpParameters, IceCandidate


def getCodecName(codec: RtpCodecParameters):
    pattern = re.compile(r'^(audio|video)/(.+)', re.I)
    match = pattern.match(codec.mimeType)
    if match:
        return match.group(2)
    else:
        raise TypeError('invalid codec.mimeType')

class MediaSection:
    def __init__(
        self,
        iceParameters: Optional[RTCIceParameters]=None,
        iceCandidates: List[IceCandidate]=[],
        dtlsParameters: Optional[RTCDtlsParameters]=None,
        planB:bool=False
    ):
        self._mediaDict: dict={}
        self._planB=planB
        if iceParameters:
            self.setIceParameters(iceParameters)
        if iceCandidates:
            self._mediaDict['candidates'] = []
            for candidate in iceCandidates:
                c_dict = candidate.dict()
                c_dict['component'] = 1
                self._mediaDict['candidates'].append(c_dict)
            self._mediaDict['endOfCandidates'] = 'end-of-candidates'
            self._mediaDict['iceOptions'] = 'renomination'
        if dtlsParameters:
            self.setDtlsRole(dtlsParameters.role)
    
    @property
    def mid(self) -> str:
        return str(self._mediaDict.get('mid', ''))
    
    @property
    def closed(self):
        return self._mediaDict.get('port') == 0
    
    def getDict(self):
        return self._mediaDict
    
    def setIceParameters(self, iceParameters: RTCIceParameters):
        self._mediaDict['iceUfrag'] = iceParameters.usernameFragment
        self._mediaDict['icePwd'] = iceParameters.password
    
    def disable(self):
        self._mediaDict['direction'] = 'inactive'
        self._mediaDict.pop('ext', None)
        self._mediaDict.pop('ssrcs', None)
        self._mediaDict.pop('ssrcGroups', None)
        self._mediaDict.pop('simulcast', None)
        self._mediaDict.pop('simulcast_03', None)
        self._mediaDict.pop('rids', None)
    
    def close(self):
        self._mediaDict['direction'] = 'inactive'
        self._mediaDict['port'] = 0
        self._mediaDict.pop('ext', None)
        self._mediaDict.pop('ssrcs', None)
        self._mediaDict.pop('ssrcGroups', None)
        self._mediaDict.pop('simulcast', None)
        self._mediaDict.pop('simulcast_03', None)
        self._mediaDict.pop('rids', None)
        self._mediaDict.pop('extmapAllowMixed', None)
    
    def setDtlsRole(self, role: str):
        logging.warning('MediaSection setDtlsRole() not implement')

    def planBReceive(self, offerRtpParameters: RtpParameters, streamId: str, trackId: str):
        logging.warning('MediaSection planBReceive() not implement')
    
    def planBStopReceiving(self, offerRtpParameters: RtpParameters):
        logging.warning('MediaSection planBStopReceiving() not implement')
    
class AnswerMediaSection(MediaSection):
    def __init__(
        self,
        offerMediaDict: dict,
        sctpParameters: Optional[SctpParameters]=None,
        offerRtpParameters: Optional[RtpParameters]=None,
        answerRtpParameters: Optional[RtpParameters]=None,
        codecOptions: Optional[ProducerCodecOptions]=None,
        iceParameters: Optional[RTCIceParameters]=None,
        iceCandidates: List[RTCIceCandidate]=[],
        dtlsParameters: Optional[RTCDtlsParameters]=None,
        plainRtpParameters: Optional[PlainRtpParameters]=None,
        planB: bool=False,
        extmapAllowMixed: bool=False
    ):
        super(AnswerMediaSection, self).__init__(iceParameters, iceCandidates, dtlsParameters, planB)
        self._mediaDict['mid'] = offerMediaDict.get('mid')
        self._mediaDict['type'] = offerMediaDict.get('type')
        self._mediaDict['protocol'] = offerMediaDict.get('protocol')

        if not plainRtpParameters:
            self._mediaDict['connection'] = {
                'ip': '127.0.0.1',
                'version': 4
            }
            self._mediaDict['port'] = 7
        else:
            self._mediaDict['connection'] = {
                'ip': plainRtpParameters.ip,
                'version': plainRtpParameters.ipVersion
            }
            self._mediaDict['port'] = plainRtpParameters.port
        
        if offerMediaDict.get('type') in ['audio', 'video']:
            self._mediaDict['direction'] = 'recvonly'
            self._mediaDict['rtp'] = []
            self._mediaDict['rtcpFb'] = []
            self._mediaDict['fmtp'] = []
            if answerRtpParameters:
                for codec in answerRtpParameters.codecs:
                    rtp = {
                        'payload': codec.payloadType,
                        'codec': getCodecName(codec),
                        'rate': codec.clockRate
                    }
                    if codec.channels:
                        if (codec.channels > 1):
                            rtp['encoding'] = codec.channels
                    self._mediaDict['rtp'].append(rtp)
                    codecParameters = codec.parameters
                    if codecOptions:
                        if offerRtpParameters:
                            offerCodecs = [offerRtpCodec for offerRtpCodec in offerRtpParameters.codecs if offerRtpCodec.payloadType == codec.payloadType]
                            if offerCodecs:
                                offerCodec:RtpCodecParameters = offerCodecs[0]
                                if codec.mimeType.lower() == 'audio/opus':
                                    if codecOptions.opusStereo != None:
                                        offerCodec.parameters['sprop-stereo'] = 1 if codecOptions.opusStereo else 0
                                        codecParameters['stereo'] = 1 if codecOptions.opusStereo else 0
                                    if codecOptions.opusFec != None:
                                        offerCodec.parameters['useinbandfec'] = 1 if codecOptions.opusFec else 0
                                        codecParameters['useinbandfec'] = 1 if codecOptions.opusFec else 0
                                    if codecOptions.opusDtx != None:
                                        offerCodec.parameters['usedtx'] = 1 if codecOptions.opusDtx else 0
                                        codecParameters['usedtx'] = 1 if codecOptions.opusDtx else 0
                                    if codecOptions.opusMaxPlaybackRate != None:
                                        codecParameters['maxplaybackrate'] = codecOptions.opusMaxPlaybackRate
                                    if codecOptions.opusMaxAverageBitrate != None:
                                        codecParameters['maxaveragebitrate'] = codecOptions.opusMaxAverageBitrate
                                    if codecOptions.opusPtime != None:
                                        offerCodec.parameters['ptime'] = codecOptions.opusPtime
                                        codecParameters['ptime'] = codecOptions.opusPtime
                                elif codec.mimeType.lower() in ['video/vp8', 'video/vp9', 'video/h264', 'video/h265']:
                                    if codecOptions.videoGoogleStartBitrate != None:
                                        codecParameters['x-google-start-bitrate'] = codecOptions.videoGoogleStartBitrate
                                    if codecOptions.videoGoogleMaxBitrate != None:
                                        codecParameters['x-google-max-bitrate'] = codecOptions.videoGoogleMaxBitrate
                                    if codecOptions.videoGoogleMinBitrate != None:
                                        codecParameters['x-google-min-bitrate'] = codecOptions.videoGoogleMinBitrate
                        
                    fmtp = {
                        'payload': codec.payloadType,
                        'config': ';'.join([f'{key}={value}' for key, value in codecParameters.items()])
                    }
                    if fmtp['config']:
                        self._mediaDict['fmtp'].append(fmtp)
                    for fb in codec.rtcpFeedback:
                        self._mediaDict['rtcpFb'].append({
                            'payload': codec.payloadType,
                            'type': fb.type,
                            'subtype': fb.parameter
                        })
                
                self._mediaDict['payloads'] = ' '.join([str(codec.payloadType) for codec in answerRtpParameters.codecs])
                self._mediaDict['ext'] = []
                for ext in answerRtpParameters.headerExtensions:
                    # Don't add a header extension if not present in the offer.
                    if ext.uri in [localExt['uri'] for localExt in offerMediaDict['ext']]:
                        self._mediaDict['ext'].append({
                            'uri': ext.uri,
                            'value': ext.id
                        })
            # Allow both 1 byte and 2 bytes length header extensions.
            if extmapAllowMixed and offerMediaDict.get('extmapAllowMixed') == 'extmap-allow-mixed':
                self._mediaDict['extmapAllowMixed'] = 'extmap-allow-mixed'
            # Simulcast.
            if offerMediaDict.get('simulcast'):
                simulcast: dict = offerMediaDict.get('simulcast', {})
                self._mediaDict['simulcast'] = {
                    'dir1': 'recv',
                    'list1': simulcast.get('list1')
                }
                self._mediaDict['rids'] = []
                if offerMediaDict.get('rids'):
                    rids: list = offerMediaDict.get('rids', [])
                    for rid in rids:
                        if rid.get('direction') == 'send':
                            self._mediaDict['rids'].append({
                                'id': rid.get('id'),
                                'direction': 'recv'
                            })
            # Simulcast (draft version 03).
            elif offerMediaDict.get('simulcast_03'):
                simulcast_03: dict = offerMediaDict.get('simulcast_03', {})
                self._mediaDict['simulcast_03'] = {
                    'value': simulcast_03.get('value', '').replace('send', 'recv')
                }
                self._mediaDict['rids'] = []
                if offerMediaDict.get('rids'):
                    rids_03: list = offerMediaDict.get('rids', [])
                    for rid in rids_03:
                        if rid.get('direction') == 'send':
                            self._mediaDict['rids'].append({
                                'id': rid.get('id'),
                                'direction': 'recv'
                            })
            self._mediaDict['rtcpMux'] = 'rtcp-mux'
            self._mediaDict['rtcpRsize'] = 'rtcp-rsize'
            if self._planB and self._mediaDict.get('type') == 'video':
                self._mediaDict['xGoogleFlag'] = 'conference'
            
        elif offerMediaDict.get('type') == 'application':
            # New spec.
            if sctpParameters:
                if offerMediaDict.get('sctpPort'):
                    self._mediaDict['payloads'] = 'webrtc-datachannel'
                    self._mediaDict['sctpPort'] = sctpParameters.port
                    self._mediaDict['maxMessageSize'] = sctpParameters.maxMessageSize
                # Old spec.
                elif offerMediaDict.get('sctpmap'):
                    self._mediaDict['payloads'] = sctpParameters.port
                    self._mediaDict['sctpmap'] = {
                        'app': 'webrtc-datachannel',
                        'sctpmapNumber': sctpParameters.port,
                        'maxMessageSize': sctpParameters.maxMessageSize
                    }
    def setDtlsRole(self, role: str):
        if role == 'client':
            self._mediaDict['setup'] = 'active'
        elif role == 'server':
            self._mediaDict['setup'] = 'passive'
        elif role == 'auto':
            self._mediaDict['setup'] = 'actpass'

class OfferMediaSection(MediaSection):
    def __init__(
        self,
        mid: str,
        kind: Literal['audio', 'video', 'application'],
        streamId: Optional[str]=None,
        trackId: Optional[str]=None,
        oldDataChannelSpec: Optional[bool]=False,
        sctpParameters: Optional[SctpParameters]=None,
        offerRtpParameters: Optional[RtpParameters]=None,
        iceParameters: Optional[RTCIceParameters]=None,
        iceCandidates: List[RTCIceCandidate]=[],
        dtlsParameters: Optional[RTCDtlsParameters]=None,
        plainRtpParameters: Optional[PlainRtpParameters]=None,
        planB: bool=False,    
    ):
        super(OfferMediaSection, self).__init__(iceParameters, iceCandidates, dtlsParameters, planB)
        self._mediaDict['mid'] = mid
        self._mediaDict['type'] = kind
        if not plainRtpParameters:
            self._mediaDict['connection'] = {
                'ip': '127.0.0.1',
                'version': 4
            }
            if not sctpParameters:
                self._mediaDict['protocol'] = 'UDP/TLS/RTP/SAVPF'
            else:
                self._mediaDict['protocol'] = 'UDP/DTLS/SCTP'
            self._mediaDict['port'] = 7
        else:
            self._mediaDict['connection'] = {
                'ip': plainRtpParameters.ip,
                'version': plainRtpParameters.ipVersion
            }
            self._mediaDict['protocol'] = 'RTP/AVP'
            self._mediaDict['port'] = plainRtpParameters.port
        
        if kind in ['audio', 'video']:
            self._mediaDict['direction'] = 'sendonly'
            self._mediaDict['rtp'] = []
            self._mediaDict['rtcpFb'] = []
            self._mediaDict['fmtp'] = []
            if not self._planB:
                self._mediaDict['msid'] = f"{streamId if streamId else '-'} {trackId}"
            if offerRtpParameters:
                for codec in offerRtpParameters.codecs:
                    rtp = {
                        'payload': codec.payloadType,
                        'codec': getCodecName(codec),
                        'rate': codec.clockRate
                    }
                    if codec.channels:
                        if codec.channels > 1:
                            rtp['encoding'] = codec.channels
                    self._mediaDict['rtp'].append(rtp)
                    fmtp = {
                        'payload': codec.payloadType,
                        'config': ';'.join([f'{key}={value}' for key, value in codec.parameters.items()])
                    }
                    if fmtp['config']:
                        self._mediaDict['fmtp'].append(fmtp)
                    for fb in codec.rtcpFeedback:
                        self._mediaDict['rtcpFb'].append({
                            'payload': codec.payloadType,
                            'type': fb.type,
                            'subtype': fb.parameter
                        })
                
                self._mediaDict['payloads'] = ' '.join([str(codec.payloadType) for codec in offerRtpParameters.codecs])
                self._mediaDict['ext'] = []
                for ext in offerRtpParameters.headerExtensions:
                    self._mediaDict['ext'].append({
                        'uri': ext.uri,
                        'value': ext.id
                    })
                
                self._mediaDict['rtcpMux'] = 'rtcp-mux'
                self._mediaDict['rtcpRsize'] = 'rtcp-rsize'
                encoding: RtpEncodingParameters = offerRtpParameters.encodings[0]
                ssrc = encoding.ssrc
                rtxSsrc = encoding.rtx.ssrc if encoding.rtx and encoding.rtx.ssrc else None
                self._mediaDict['ssrcs'] = []
                self._mediaDict['ssrcGroups'] = []
                cname = None
                if offerRtpParameters.rtcp:
                    if offerRtpParameters.rtcp.cname:
                        cname = offerRtpParameters.rtcp.cname
                if cname:
                    self._mediaDict['ssrcs'].append({
                        'id': ssrc,
                        'attribute': 'cname',
                        'value': cname
                    })
                if self._planB:
                    self._mediaDict['ssrcs'].append({
                        'id': ssrc,
                        'attribute': 'msid',
                        'value': f"{streamId if streamId else '-'} {trackId}"
                    })
                if rtxSsrc:
                    if cname:
                        self._mediaDict['ssrcs'].append({
                            'id': rtxSsrc,
                            'attribute': 'cname',
                            'value': cname
                        })
                    if self._planB:
                        self._mediaDict['ssrcs'].append({
                            'id': rtxSsrc,
                            'attribute': 'msid',
                            'value': f"{streamId if streamId else '-'} {trackId}"
                        })
                    self._mediaDict['ssrcGroups'].append({
                        'semantics': 'FID',
                        'ssrcs': f'{ssrc} {rtxSsrc}'
                    })
        elif kind == 'application':
            if sctpParameters:
                # New spec.
                if not oldDataChannelSpec:
                    self._mediaDict['payloads'] = 'webrtc-datachannel'
                    self._mediaDict['sctpPort'] = sctpParameters.port
                    self._mediaDict['maxMessageSize'] = sctpParameters.maxMessageSize
                # Old spec.
                else:
                    self._mediaDict['payloads'] = sctpParameters.port
                    self._mediaDict['sctpmap'] = {
                        'app': 'webrtc-datachannel',
                        'sctpmapNumber': sctpParameters.port,
                        'maxMessageSize': sctpParameters.maxMessageSize
                    }
    def setDtlsRole(self, _):
        # Always 'actpass'.
        self._mediaDict['setup'] = 'actpass'
    
    def planBReceive(self, offerRtpParameters: RtpParameters, streamId: str, trackId: str):
        encoding = offerRtpParameters.encodings[0]
        ssrc = encoding.ssrc
        rtxSsrc = encoding.rtx.ssrc if encoding.rtx and encoding.rtx.ssrc else None
        if offerRtpParameters.rtcp:
            if offerRtpParameters.rtcp.cname:
                cname = offerRtpParameters.rtcp.cname
        if cname:
            self._mediaDict['ssrcs'].append({
                'id': ssrc,
                'attribute': 'cname',
                'value': cname
            })
        self._mediaDict['ssrcs'].append({
            'id': ssrc,
            'attribute': 'msid',
            'value': f"{streamId if streamId else '-'} {trackId}"
        })
        if rtxSsrc:
            if cname:
                self._mediaDict['ssrcs'].append({
                    'id': rtxSsrc,
                    'attribute': 'cname',
                    'value': cname
                })
            self._mediaDict['ssrcs'].append({
                'id': rtxSsrc,
                'attribute': 'msid',
                'value': f"{streamId if streamId else '-'} {trackId}"
            })
            self._mediaDict['ssrcGroups'].append({
                'semantics': 'FID',
                'ssrcs': f'{ssrc} {rtxSsrc}'
            })
    
    def planBStopReceiving(self, offerRtpParameters: RtpParameters):
        encoding = offerRtpParameters.encodings[0]
        ssrc = encoding.ssrc
        rtxSsrc = encoding.rtx.ssrc if encoding.rtx and encoding.rtx.ssrc else None
        self._mediaDict['ssrcs'] = [s for s in self._mediaDict['ssrcs'] if s['id'] != ssrc and s['id'] != rtxSsrc]
        if rtxSsrc:
            self._mediaDict['ssrcGroups'] = [group for group in self._mediaDict['ssrcGroups'] if group['ssrcs'] != f'{ssrc} {rtxSsrc}']



from .api.mediasoup_client import MediasoupClient
from .api.mediasoup_signaler import MediasoupSignalerInterface
from .api.notification_listener import BandwidthNotificationListener, PeerNotificationListener, \
    ProducerNotificationListener, ConsumerNotificationListener, DataConsumerNotificationListener
from .api.request_listener import ConsumerRequestListener, DataConsumerRequestListener
from .device import Device
from .handlers.aiortc_handler import AiortcHandler



# h264-profile-level-id

# Profile
ProfileConstrainedBaseline = 1
ProfileBaseline = 2
ProfileMain = 3
ProfileConstrainedHigh = 4
ProfileHigh = 5

# Level
# All values are equal to ten times the level number, except level 1b which is
# special.
Level1_b = 0
Level1 = 10
Level1_1 = 11
Level1_2 = 12
Level1_3 = 13
Level2 = 20
Level2_1 = 21
Level2_2 = 22
Level3 = 30
Level3_1 = 31
Level3_2 = 32
Level4 = 40
Level4_1 = 41
Level4_2 = 42
Level5 = 50
Level5_1 = 51
Level5_2 = 52


class ProfileLevelId:
    def __init__(self, profile, level):
        self.profile = profile
        self.level = level

# Default ProfileLevelId.
#
# TODO: The default should really be profile Baseline and level 1 according to
# the spec: https://tools.ietf.org/html/rfc6184#section-8.1. In order to not
# break backwards compatibility with older versions of WebRTC where external
# codecs don't have any parameters, use profile ConstrainedBaseline level 3_1
# instead. This workaround will only be done in an interim period to allow
# external clients to update their code.
#
# http://crbug/webrtc/6337.
DefaultProfileLevelId = ProfileLevelId(ProfileConstrainedBaseline, Level3_1)

# For level_idc=11 and profile_idc=0x42, 0x4D, or 0x58, the constraint set3
# flag specifies if level 1b or level 1.1 is used.
ConstraintSet3Flag = 0x10

# Convert a string of 8 characters into a byte where the positions containing
# character c will have their bit set. For example, c = 'x', str = "x1xx0000"
# will return 0b10110000.
def byteMaskString(c, str):
    return (
        ((str[0] == c) << 7) | ((str[1] == c) << 6) | ((str[2] == c) << 5) |
        ((str[3] == c) << 4) | ((str[4] == c) << 3)    | ((str[5] == c) << 2)    |
        ((str[6] == c) << 1) | ((str[7] == c) << 0)
    )


# Class for matching bit patterns such as "x1xx0000" where 'x' is allowed to be
# either 0 or 1.
class BitPattern:
    def __init__(self, pattern):
        self._mask = ~byteMaskString('x', pattern)
        self._maskedValue = byteMaskString('1', pattern)
    
    def isMatch(self, value):
        return self._maskedValue == (value & self._mask)

class ProfilePattern:
    def __init__(self, profile_idc, profile_iop, profile):
        self.profile_idc = profile_idc
        self.profile_iop = profile_iop
        self.profile = profile

ProfilePatterns = [
    ProfilePattern(0x42, BitPattern('x1xx0000'), ProfileConstrainedBaseline),
    ProfilePattern(0x4D, BitPattern('1xxx0000'), ProfileConstrainedBaseline),
	ProfilePattern(0x58, BitPattern('11xx0000'), ProfileConstrainedBaseline),
	ProfilePattern(0x42, BitPattern('x0xx0000'), ProfileBaseline),
	ProfilePattern(0x58, BitPattern('10xx0000'), ProfileBaseline),
	ProfilePattern(0x4D, BitPattern('0x0x0000'), ProfileMain),
	ProfilePattern(0x64, BitPattern('00000000'), ProfileHigh),
	ProfilePattern(0x64, BitPattern('00001100'), ProfileConstrainedHigh)
]

# Parse profile level id that is represented as a string of 3 hex bytes.
# Nothing will be returned if the string is not a recognized H264 profile
# level id.
#
# @param {String} str - profile-level-id value as a string of 3 hex bytes.
#
# @returns {ProfileLevelId}
def parseProfileLevelId(level_id: str = None):
    if not level_id:
        return None
    
    if len(level_id) != 6:
        return None

    try:
        profile_level_id_numeric = int(level_id, 16)
    except ValueError:
        return None

    if profile_level_id_numeric == 0:
        return None
    
    level_idc = profile_level_id_numeric & 0xFF
    profile_iop = (profile_level_id_numeric >> 8) & 0xFF
    profile_idc = (profile_level_id_numeric >> 16) & 0xFF

    if (level_idc == Level1_1):
        level = Level1_b if (profile_iop & ConstraintSet3Flag) != 0 else Level1_1
    elif level_idc in [Level1, Level1_2, Level1_3, Level2, Level2_1, Level2_2, Level3, Level3_1, Level3_2, Level4, Level4_1, Level4_2, Level5, Level5_1, Level5_2]:
        level = level_idc
    else:
        return None
    
    # Parse profile_idc/profile_iop into a Profile enum.
    for pattern in ProfilePatterns:
        if profile_idc == pattern.profile_idc and pattern.profile_iop.isMatch(profile_iop):
            return ProfileLevelId(pattern.profile, level)
    
    return None

# Returns canonical string representation as three hex bytes of the profile
# level id, or returns nothing for invalid profile level ids.
#
# @param {ProfileLevelId} profile_level_id
#
# @returns {String}
def profileLevelIdToString(profile_level_id):
    if profile_level_id.level == Level1_b:
        if profile_level_id.profile == ProfileConstrainedBaseline:
            return '42f00b'
        elif profile_level_id.profile == ProfileBaseline:
            return '42100b'
        elif profile_level_id.profile == ProfileMain:
            return '4d100b'
        else:
            return None
    
    if profile_level_id.profile == ProfileConstrainedBaseline:
        profile_idc_iop_string = '42e0'
    elif profile_level_id.profile == ProfileBaseline:
        profile_idc_iop_string = '4200'
    elif profile_level_id.profile == ProfileMain:
        profile_idc_iop_string = '4d00'
    elif profile_level_id.profile == ProfileConstrainedHigh:
        profile_idc_iop_string = '640c'
    elif profile_level_id.profile == ProfileHigh:
        profile_idc_iop_string = '6400'
    else:
        return None
    
    return f'{profile_idc_iop_string}{profile_level_id.level:02x}'

# Parse profile level id that is represented as a string of 3 hex bytes
# contained in an SDP key-value map. A default profile level id will be
# returned if the profile-level-id key is missing. Nothing will be returned if
# the key is present but the string is invalid.
#
# @param {Object} [params={}] - Codec parameters object.
#
# @returns {ProfileLevelId}
def parseSdpProfileLevelId(params={}):
    profile_level_id = params.get('profile-level-id')
    return DefaultProfileLevelId if profile_level_id == None else parseProfileLevelId(profile_level_id)

# Returns True if the parameters have the same H264 profile, i.e. the same
# H264 profile (Baseline, High, etc).
#
# @param {Object} [params1={}] - Codec parameters object.
# @param {Object} [params2={}] - Codec parameters object.
#
# @returns {Boolean}
def isSameProfile(params1={}, params2={}):
    profile_level_id_1 = parseSdpProfileLevelId(params1)
    profile_level_id_2 = parseSdpProfileLevelId(params2)
    return profile_level_id_1 and profile_level_id_2 and (profile_level_id_1.profile == profile_level_id_2.profile)

# Generate codec parameters that will be used as answer in an SDP negotiation
# based on local supported parameters and remote offered parameters. Both
# local_supported_params and remote_offered_params represent sendrecv media
# descriptions, i.e they are a mix of both encode and decode capabilities. In
# theory, when the profile in local_supported_params represent a strict superset
# of the profile in remote_offered_params, we could limit the profile in the
# answer to the profile in remote_offered_params.
#
# However, to simplify the code, each supported H264 profile should be listed
# explicitly in the list of local supported codecs, even if they are redundant.
# Then each local codec in the list should be tested one at a time against the
# remote codec, and only when the profiles are equal should this function be
# called. Therefore, this function does not need to handle profile intersection,
# and the profile of local_supported_params and remote_offered_params must be
# equal before calling this function. The parameters that are used when
# negotiating are the level part of profile-level-id and level-asymmetry-allowed.
#
# @param {Object} [local_supported_params={}]
# @param {Object} [remote_offered_params={}]
#
# @returns {String} Canonical string representation as three hex bytes of the
#   profile level id, or null if no one of the params have profile-level-id.
#
# @throws {TypeError} If Profile mismatch or invalid params.
def generateProfileLevelIdForAnswer(local_supported_params={},remote_offered_params={}):
    if not local_supported_params.get('profile-level-id') and not remote_offered_params.get('profile-level-id'):
        return None
    
    local_profile_level_id = parseSdpProfileLevelId(local_supported_params)
    remote_profile_level_id = parseSdpProfileLevelId(remote_offered_params)

    # The local and remote codec must have valid and equal H264 Profiles.
    if not local_profile_level_id:
        raise TypeError('invalid local_profile_level_id')
    
    if not remote_profile_level_id:
        raise TypeError('invalid remote_profile_level_id')

    if local_profile_level_id.profile != remote_profile_level_id.profile:
        raise TypeError('H264 Profile mismatch')

    level_asymmetry_allowed = isLevelAsymmetryAllowed(local_supported_params) and isLevelAsymmetryAllowed(remote_offered_params)

    local_level = local_profile_level_id.level
    remote_level = remote_profile_level_id.level
    min_level = minLevel(local_level, remote_level)
    answer_level = local_level if level_asymmetry_allowed else min_level

    return profileLevelIdToString(ProfileLevelId(local_profile_level_id.profile, answer_level))

# Compare H264 levels and handle the level 1b case.
def isLessLevel(a, b):
    if a == Level1_b:
        return b != Level1 and b != Level1_b
    if b == Level1_b:
        return a != Level1
    return a < b

def minLevel(a, b):
    return a if isLessLevel(a, b) else b

def isLevelAsymmetryAllowed(params = {}):
    level_asymmetry_allowed = params.get('level-asymmetry-allowed')
    return level_asymmetry_allowed == 1 or level_asymmetry_allowed == '1'


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['smcdk',
 'smcdk.api',
 'smcdk.deps.h264_profile_level_id',
 'smcdk.deps.sdp_transform',
 'smcdk.handlers',
 'smcdk.handlers.sdp',
 'smcdk.log',
 'smcdk.models']

package_data = \
{'': ['*'], 'smcdk': ['deps/*']}

install_requires = \
['aiortc>=1.2.0,<2.0.0', 'pydantic>=1.8.1,<2.0.0', 'pyee>=8.1.0,<9.0.0']

setup_kwargs = {
    'name': 'smcdk',
    'version': '1.0.0',
    'description': "mediasoup-client-pysdk, a simple mediasoup client development kit, i.e. 'smcdk', fork from pymediasoup and do more",
    'long_description': '# Mediasoup-Client-PySDK (aka "smcdk")\na simple-to-use, pure python sdk of [mediasoup](https://mediasoup.org/) client, fork from [pymediasoup](https://github.com/skymaze/pymediasoup) and do more.\n\n## Usage\nFor the purpose of to be an easy-to-use SDK, smcdk API design focus mainly on high level, and for users who know little about the official mediasoup client API. \n\n```python\nfrom smcdk import *\n\nmediasoup_client = MediasoupClient(...)\nmediasoup_client.joinRoom(...)\nmediasoup_client.play(...)\nmediasoup_client.close()\n```\nmore details, please see: examples/sdkApiDemo.py\n\n## Why another mediasoup-client?( My Personal Option, for reference only)\nThere are several official and unofficial client implementations, but they are not quick and easy to run on all OS\'s desktop, so are not suitable to be a general SDK: \n1. official client with official dependency lib\n- mediasoup-demo/aiortc: because it is based on Unix Socket, so it can\'t run in Windows\n- mediasoup-demo/broadcasters: it is based on bash language, which is good at integrating command line tools, but is not good at developing new features\n- mediasoup-demo/app: it can only run in browsers, and Electron-like desktop environment with less disk space occupation, or run in Node.js with more space occupation because of the node_modules directory\n- mediasoup-broadcast-demo: it\'s quite hard to compile and link a libwebrtc dependency successfully on all OS platform, especially in China mainland\'s network environment\n\n2. no-official client\n- pymediasoup： it is quite nice, but its API is a little hard to quick start as SDK\n\n## Architecture & Design\n![image](resources/architecture.png)\n\n### Mediasoup Client\nit contains:\n- Mediasoup Signaler Interface: follow the semantics of mediasoup-demo/server\n- Loop Tasks & Listeners: to tackle signaler request and notification from server side\n- Room and Peer: a group of simple APIs to be integrated to Listeners\n- Multimedia Runtime: a stateful mediasoup Device\n\n### Business Domain Based Listener Design\nThere are several business domain in SDK design:\nBandwidth, Peer, Producer, Consumer, DataConsumer, result in 2 request listeners\nand 5 notification listeners, which their Respective interesting events to listen and tackle\n1. Server Request\n- Consumer Listener event: newConsumer\n- DataConsumer Listener event: newDataConsumer\n2. Server Notification\n- Bandwidth Listener event: downlinkBwe\n- Peer Listener event: newPeer, peerClosed, peerDisplayNameChanged, activeSpeaker\n- Producer Listener event: producerScore\n- Consumer Listener event: consumerLayersChanged, consumerScore, consumerClosed, consumerPaused, consumerResumed\n- DataConsumer Listener event: dataConsumerClosed\n\n## Features\nTo be an easy-to-use sdk for mediasoup client development\n- **quick to run**: as mentioned above\n- **all os platform friendly**: as mentioned above\n- **signaling pluggable**: based on the mediasoup\'s design goal of "signaling agnostic", \n    >Be signaling agnostic: do not mandate any signaling protocol.”\n   \n    (sited from [mediasoup :: Overview](https://mediasoup.org/documentation/overview/)). \n    smcdk provide an out-of-box ProtooSignaler furthermore. \n    Besides the default built-in signaler, which is used to communicate with mediasoup-demo/server, \n    you can provide your own MediasoupSignalerInterface implementation to meet your requirement.\n- **listener customizable**: currently, you can customize 2 request listeners and 5 notification listeners \n## About Code Style\n  You can see many Camel-Case-Style naming in my Python code, \ne.g. "getRouterRtpCapabilities", not "get_router_rtp_capabilities".\n  The reason is not only that I began my career as a Java developer since 2008,\nbut also that I hope this SDK can be applied by those developers who use Python as a no-major language, \nand developers who have learned mediasoup by its demo app & server.\n  Maybe sometime in the future, I\'ll change this naming to follow Python’s PEP8 rules.\n## LICENSE\nMIT\n\n## Thanks\nspecial thanks to pymediasoup, mediasoup, and aiortc projects\n',
    'author': 'Jim DG',
    'author_email': 'jimdragongod@126.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/jimdragongod/mediasoup-client-pysdk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.0,<4.0.0',
}


setup(**setup_kwargs)
